#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""PySpectral configuration directory and file handling."""

import logging
import os
from collections.abc import Mapping
from os.path import expanduser
from pathlib import Path

import yaml
from platformdirs import AppDirs

try:
    from yaml import UnsafeLoader
except ImportError:
    from yaml import Loader as UnsafeLoader


LOG = logging.getLogger(__name__)

BUILTIN_CONFIG_FILE = Path(__file__).resolve().parent / "etc" / "pyspectral.yaml"


def recursive_dict_update(d, u):
    """Recursive dictionary update.

    Copied from:

        http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth

    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = recursive_dict_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def get_config(config_file: str | Path = None) -> dict:
    """Get configuration options from YAML file."""
    if config_file is None:
        config_file = _get_env_or_builtin_config_path()

    config = {}
    with open(config_file, 'r') as fp_:
        config = recursive_dict_update(config, yaml.load(fp_, Loader=UnsafeLoader))

    app_dirs = AppDirs('pyspectral', 'pytroll')
    user_datadir = app_dirs.user_data_dir
    config['rsr_dir'] = expanduser(config.get('rsr_dir', user_datadir))
    config['rayleigh_dir'] = expanduser(config.get('rayleigh_dir', user_datadir))
    os.makedirs(config['rsr_dir'], exist_ok=True)
    os.makedirs(config['rayleigh_dir'], exist_ok=True)

    return config


def _get_env_or_builtin_config_path() -> Path:
    config_file = os.environ.get('PSP_CONFIG_FILE')
    if config_file is not None and not os.path.isfile(config_file):
        raise IOError(f"{config_file} pointed to by the environment variable "
                      f"'PSP_CONFIG_FILE' is not a file or does not exist!")
    if config_file is None:
        return BUILTIN_CONFIG_FILE
    return Path(config_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 - 2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Download the atmospheric corrections Look-Up Tables.

Script to download the the LUT files for atmospheric correction in the Short
Wave spectral range.

"""

import argparse
import logging

from pyspectral.rayleigh import check_and_download
from pyspectral.utils import AEROSOL_TYPES, logging_off, logging_on

LOG = logging.getLogger(__name__)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Download the atm correction LUT files')
    parser.add_argument("--aerosol_types", '-a', nargs='*',
                        help="Aerosol types",
                        type=str, default=AEROSOL_TYPES)
    parser.add_argument(
        "-d", '--dry_run', help=("Dry run - no action"), action='store_true',
        default=False)
    parser.add_argument(
        "-v", '--verbose', help=("Turn logging on"), action='store_true')

    args = parser.parse_args()
    verbose = args.verbose
    dry_run = args.dry_run
    aerosol_types = args.aerosol_types

    if verbose:
        logging_on(logging.DEBUG)
    else:
        logging_off()

    check_and_download(aerosol_types=aerosol_types, dry_run=dry_run)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2023 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Planck radiation equation."""

import logging

import numpy as np

try:
    import dask.array as da
except ImportError:
    da = np

from pyspectral.utils import use_map_blocks_on

LOG = logging.getLogger(__name__)

H_PLANCK = 6.62606957 * 1e-34  # SI-unit = [J*s]
K_BOLTZMANN = 1.3806488 * 1e-23  # SI-unit = [J/K]
C_SPEED = 2.99792458 * 1e8  # SI-unit = [m/s]

PLANCK_C1 = H_PLANCK * C_SPEED / K_BOLTZMANN
PLANCK_C2 = 2 * H_PLANCK * C_SPEED**2

EPSILON = 0.000001


@use_map_blocks_on("radiance")
def blackbody_rad2temp(wavelength, radiance):
    """Derive brightness temperatures from radiance using the inverse Planck function.

    Args:
        wavelength: The wavelength to use, in SI units (metre).
        radiance: The radiance to derive temperatures from, in SI units (W/m^2 sr^-1). Scalar or arrays are accepted.

    Returns:
        The derived temperature in Kelvin.

    """
    if getattr(wavelength, "dtype", None) != radiance.dtype:
        # avoid a wavelength numpy scalar upcasting radiances (ex. 32-bit to 64-bit float)
        wavelength = radiance.dtype.type(wavelength)
    with np.errstate(invalid='ignore'):
        return PLANCK_C1 / (wavelength * np.log(PLANCK_C2 / (radiance * wavelength**5) + 1.0))


@use_map_blocks_on("radiance")
def blackbody_wn_rad2temp(wavenumber, radiance):
    """Derive brightness temperatures from radiance using the inverse Planck function.

    Args:
        wavenumber: The wavenumber to use, in SI units (1/metre).
        radiance: The radiance to derive temperatures from, in SI units (W/m^2 sr^-1). Scalar or arrays are accepted.

    Returns:
        The derived temperature in Kelvin.

    """
    with np.errstate(invalid='ignore'):
        return PLANCK_C1 * wavenumber / np.log((PLANCK_C2 * wavenumber**3) / radiance + 1.0)


def planck(wave, temperature, wavelength=True):
    """Derive the Planck radiation as a function of wavelength or wavenumber.

    SI units.
    _planck(wave, temperature, wavelength=True)
    wave = Wavelength/wavenumber or a sequence of wavelengths/wavenumbers (m or m^-1)
    temp = Temperature (scalar) or a sequence of temperatures (K)


    Output: Wavelength space: The spectral radiance per meter (not micron!)
            Unit = W/m^2 sr^-1 m^-1

            Wavenumber space: The spectral radiance in Watts per square meter
            per steradian per m-1:
            Unit = W/m^2 sr^-1 (m^-1)^-1 = W/m sr^-1

            Converting from SI units to mW/m^2 sr^-1 (cm^-1)^-1:
            1.0 W/m^2 sr^-1 (m^-1)^-1 = 1.0e5 mW/m^2 sr^-1 (cm^-1)^-1

    """
    units = ['wavelengths', 'wavenumbers']
    if wavelength:
        LOG.debug("Using {0} when calculating the Blackbody radiance".format(
            units[(wavelength is True) - 1]))

    if wavelength:
        nom = PLANCK_C2 / wave ** 5
        arg1 = PLANCK_C1 / wave
    else:
        nom = PLANCK_C2 * wave ** 3
        arg1 = PLANCK_C1 * wave

    with np.errstate(divide='ignore', invalid='ignore'):
        # use dask functions when needed
        np_ = np if isinstance(temperature, np.ndarray) else da
        arg2 = np_.where(np_.greater(np.abs(temperature), EPSILON),
                         np_.divide(1., temperature), np.nan).reshape(-1, 1)

    if isinstance(arg2, np.ndarray):
        # don't compute min/max if we have dask arrays
        LOG.debug("Max and min - arg1: %s  %s",
                  str(np.nanmax(arg1)), str(np.nanmin(arg1)))
        LOG.debug("Max and min - arg2: %s  %s",
                  str(np.nanmax(arg2)), str(np.nanmin(arg2)))

    try:
        exp_arg = np.multiply(arg1, arg2)
    except MemoryError:
        LOG.warning(("Dimensions used in numpy.multiply probably reached "
                     "limit!\n"
                     "Make sure the Radiance<->Tb table has been created "
                     "and try running again"))
        raise

    if isinstance(exp_arg, np.ndarray) and exp_arg.min() < 0:
        LOG.debug("Max and min before exp: %s  %s",
                  str(exp_arg.max()), str(exp_arg.min()))
        LOG.warning("Something is fishy: \n" +
                    "\tDenominator might be zero or negative in radiance derivation:")
        dubious = np.where(exp_arg < 0)[0]
        LOG.warning(
            "Number of items having dubious values: " + str(dubious.shape[0]))

    with np.errstate(over='ignore'):
        denom = np.exp(exp_arg) - 1
        return nom / denom


def blackbody_wn(wavenumber, temp):
    """Derive the Planck radiation as a function of wavenumber.

    SI units.
    blackbody_wn(wavnum, temperature)
    wavenumber = A wavenumber (scalar) or a sequence of wave numbers (m-1)
    temp = A temperatfure (scalar) or a sequence of temperatures (K)

    Output: The spectral radiance in Watts per square meter per steradian
            per m-1:
            Unit = W/m^2 sr^-1 (m^-1)^-1 = W/m sr^-1

            Converting from SI units to mW/m^2 sr^-1 (cm^-1)^-1:
            1.0 W/m^2 sr^-1 (m^-1)^-1 = 1.0e5 mW/m^2 sr^-1 (cm^-1)^-1

    """
    return planck(wavenumber, temp, wavelength=False)


def blackbody(wavel, temp):
    """Derive the Planck radiation as a function of wavelength.

    SI units.
    blackbody(wavelength, temperature)
    wavel = Wavelength or a sequence of wavelengths (m)
    temp = Temperature (scalar) or a sequence of temperatures (K)

    Output: The spectral radiance per meter (not micron!)
            Unit = W/m^2 sr^-1 m^-1

    """
    return planck(wavel, temp, wavelength=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Converting between radiance and Tb.

Conversion between radiances and brightness temperatures for the IR bands of
various satellite sensors
"""

import logging
from numbers import Number

import numpy as np

from pyspectral.blackbody import C_SPEED, H_PLANCK, K_BOLTZMANN, blackbody, blackbody_wn
from pyspectral.rsr_reader import RelativeSpectralResponse
from pyspectral.utils import BANDNAMES, WAVE_LENGTH, WAVE_NUMBER, convert2wavenumber, get_bandname_from_wavelength

try:
    from scipy.integrate import trapezoid
except ImportError:
    from scipy.integrate import trapz as trapezoid

LOG = logging.getLogger(__name__)

BLACKBODY_FUNC = {WAVE_LENGTH: blackbody,
                  WAVE_NUMBER: blackbody_wn}

EPSILON = 0.01
TB_MIN = 150.
TB_MAX = 360.

# Meteosat SEVIRI regression parameters according to documentation
# (PDF_EFFECT_RAD_TO_BRIGHTNESS.pdf).
#
# Tb = C2 * νc/{α * log[C1*νc**3 / L + 1]} - β/α
#
# L = C1 * νc**3 / (exp (C2 νc / [αTb + β]) − 1)
#
# C1 = 2 * h * c**2 and C2 = hc/k
#
# Units are cm-1 for the channel/band central wavenumber, K for the beta
# parameter, and the alpha parameter is dimensionless:
#
SEVIRI = {'IR3.9': {'Meteosat-8': [2567.330, 0.9956, 3.410],
                    'Meteosat-9': [2568.832, 0.9954, 3.438],
                    'Meteosat-10': [],
                    },
          'WV6.2': {'Meteosat-8': [1598.103, 0.9962, 2.218],
                    'Meteosat-9': [1600.548, 0.9963, 2.185],
                    },
          'WV7.3': {'Meteosat-8': [1362.081, 0.9991, 0.478],
                    'Meteosat-9': [1360.330, 0.9991, 0.470],
                    },
          'IR8.7': {'Meteosat-8': [1149.069, 0.9996, 0.179],
                    'Meteosat-9': [1148.620, 0.9996, 0.179],
                    },
          'IR9.7': {'Meteosat-8': [1034.343, 0.9999, 0.060],
                    'Meteosat-9': [1035.289, 0.9999, 0.056],
                    },
          'IR10.8': {'Meteosat-8': [930.647, 0.9983, 0.625],
                     'Meteosat-9': [931.700, 0.9983, 0.640],
                     },
          'IR12.0': {'Meteosat-8': [839.660, 0.9988, 0.397],
                     'Meteosat-9': [836.445, 0.9988, 0.408],
                     },
          'IR13.4': {'Meteosat-8': [752.387, 0.9981, 0.578],
                     'Meteosat-9': [751.792, 0.9981, 0.561],
                     },
          }


class RadTbConverter(object):
    """A radiance to brightness temperature calculator.

    It does the conversion based on direct use of the band relative
    spectral response functions.

    """

    def __init__(self, platform_name, instrument, band, detector='det-1', wavespace=WAVE_LENGTH,
                 tb_resolution=0.1):
        """Initialize the Class instance.

        E.g.:
        platform_name = 'Meteosat-9'
        instrument = 'seviri'
        band = 3.75

        """
        self.platform_name = platform_name
        self.instrument = instrument
        self.response = None
        self.wavelength_or_wavenumber = None
        self.bandname = None
        self.bandwavelength = None
        self.band = band

        self.wavespace = wavespace
        if self.wavespace not in [WAVE_LENGTH, WAVE_NUMBER]:
            raise AttributeError('Wave space not {0} or {1}!'.format(WAVE_LENGTH,
                                                                     WAVE_NUMBER))

        self._wave_unit = 'm'
        self._wave_si_scale = 1.0

        self.detector = detector
        self.tb_resolution = tb_resolution
        self.tb_scale = 1. / self.tb_resolution

        self.blackbody_function = BLACKBODY_FUNC[self.wavespace]
        self.rsr_integral = 1.0

        self._get_rsr()

    def _get_rsr(self):
        """Get the relative spectral responses.

        Get the relative spectral responses from file, find the bandname, and
        convert to the requested wave-spave (wavelength or wave number)

        """
        sensor = RelativeSpectralResponse(self.platform_name, self.instrument)

        if self.wavespace == WAVE_NUMBER:
            LOG.debug("Converting to wavenumber...")
            self.rsr, info = convert2wavenumber(sensor.rsr)
        else:
            self.rsr = sensor.rsr
            info = {'unit': sensor.unit, 'si_scale': sensor.si_scale}

        self._wave_unit = info['unit']
        self._wave_si_scale = info['si_scale']

        if isinstance(self.band, str):
            self.bandname = BANDNAMES.get(self.instrument, BANDNAMES['generic']).get(self.band, self.band)
        elif isinstance(self.band, Number):
            self.bandwavelength = self.band
            self.bandname = get_bandname_from_wavelength(self.instrument, self.band, self.rsr)

        self.wavelength_or_wavenumber = (self.rsr[self.bandname][self.detector][self.wavespace] *
                                         self._wave_si_scale)
        self.response = self.rsr[self.bandname][self.detector]['response']
        # Get the integral of the spectral response curve:
        self.rsr_integral = np.trapz(self.response, self.wavelength_or_wavenumber)

    def _getsatname(self):
        """Get the satellite name used in the rsr-reader, from the platform and number."""
        if not self.platform_name.startswith("Meteosat"):
            raise NotImplementedError(
                'Platform {0} not yet supported...'.format(self.platform_name))

        return self.platform_name

    @staticmethod
    def _getitem(block, lut):
        """Index the lut, numpy style."""
        return lut[block]

    def tb2radiance(self, tb_, **kwargs):
        """Get the radiance from the brightness temperature (Tb) given the band name.

        Input:
          tb_: Brightness temperature of the band (self.band)

        Optional arguments:
          lut: If not none, this is a Look Up Table with tb and radiance values
            which will be used for the conversion. Default is None.

          normalized: If True, the derived radiance values are the spectral radiances for the band.
            If False the radiance is the band integrated radiance. Default is True.

        """
        lut = kwargs.get('lut', None)
        normalized = kwargs.get('normalized', True)

        if self.wavespace == WAVE_NUMBER:
            if normalized:
                unit = 'W/m^2 sr^-1 (m^-1)^-1'
            else:
                unit = 'W/m^2 sr^-1'
            scale = 1.0
        else:
            if normalized:
                unit = 'W/m^2 sr^-1 m^-1'
            else:
                unit = 'W/m^2 sr^-1'
            scale = 1.0

        if lut:
            lut_radiance = lut['radiance'].astype(tb_.dtype)
            ntb = (tb_ * self.tb_scale).astype('int16')
            start = int(lut['tb'][0] * self.tb_scale)
            retv = {}
            bounds = 0, lut_radiance.shape[0] - 1
            index = (ntb - start).clip(bounds[0], bounds[1])
            try:
                retv['radiance'] = index.map_blocks(self._getitem, lut_radiance, dtype=lut_radiance.dtype)
            except AttributeError:
                retv['radiance'] = lut_radiance[index]
            try:
                retv['radiance'] = retv['radiance'].item()
            except (ValueError, AttributeError):
                pass
            retv['unit'] = unit
            retv['scale'] = scale
            return retv

        planck = self.blackbody_function(self.wavelength_or_wavenumber, tb_) * self.response
        if normalized:
            radiance = trapezoid(planck, self.wavelength_or_wavenumber) / self.rsr_integral
        else:
            radiance = trapezoid(planck, self.wavelength_or_wavenumber)

        return {'radiance': radiance,
                'unit': unit,
                'scale': scale}

    def make_tb2rad_lut(self, filepath, normalized=True):
        """Generate a Tb to radiance look-up table."""
        tb_ = np.arange(TB_MIN, TB_MAX, self.tb_resolution)
        retv = self.tb2radiance(tb_, normalized=normalized)
        rad = retv['radiance']
        np.savez(filepath, tb=tb_, radiance=rad)

    @staticmethod
    def read_tb2rad_lut(filepath):
        """Read the Tb to radiance look-up table."""
        retv = np.load(filepath, 'r')
        return retv

    def radiance2tb(self, rad):
        """Get the Tb from the radiance using the Planck function and the central wavelength of the band.

        rad:
            Radiance in SI units
        """
        return radiance2tb(rad, self.rsr[self.bandname][self.detector]['central_wavelength'] * 1e-6)


def radiance2tb(rad, wavelength):
    """Get the Tb from the radiance using the Planck function.

    rad:
        Radiance in SI units
    wavelength:
        Wavelength in SI units (meter)
    """
    from pyspectral.blackbody import blackbody_rad2temp as rad2temp
    return rad2temp(wavelength, rad)


class SeviriRadTbConverter(RadTbConverter):
    """
    Radiance<->Tb converter for SEVIRI.

    A radiance to brightness temperature calculator for SEVIRI based on
    tabulated standard values using non-linear regression methods, and thus no
    use of off line relative spectral response functions
    """

    def __init__(self, platform_name, band, **kwargs):
        """Initialize the Class instance.

        E.g.:
        platform_name = Meteosat-9
        band = 3.75

        """
        super(SeviriRadTbConverter, self).__init__(platform_name, 'seviri',
                                                   band, **kwargs)

        if isinstance(self.band, str):
            self.bandname = BANDNAMES.get(self.instrument, BANDNAMES['generic']).get(self.band, self.band)
        else:
            raise AttributeError('Band name provided as a string is required')

    def _get_rsr(self):
        """Overload the _get_rsr method, since RSR data are ignored here."""
        LOG.debug("RSR data are ignored in this converter!")

    def radiance2tb(self, rad):
        """Get the Tb from the radiance using the simple non-linear regression method.

        rad: Radiance in units = 'mW/m^2 sr^-1 (cm^-1)^-1'

        """
        #
        # Tb = C2 * νc/{α * log[C1*νc**3 / L + 1]} - β/α
        #
        # C1 = 2 * h * c**2 and C2 = hc/k
        #
        c_1 = 2 * H_PLANCK * C_SPEED ** 2
        c_2 = H_PLANCK * C_SPEED / K_BOLTZMANN

        vc_ = SEVIRI[self.bandname][self.platform_name][0]
        # Multiply by 100 to get SI units!
        vc_ *= 100.0
        alpha = SEVIRI[self.bandname][self.platform_name][1]
        beta = SEVIRI[self.bandname][self.platform_name][2]

        tb_ = c_2 * vc_ / \
            (alpha * np.log(c_1 * vc_ ** 3 / rad + 1)) - beta / alpha

        return tb_

    def tb2radiance(self, tb_, **kwargs):
        """Get the radiance from the Tb using the simple non-linear regression method.

        SI units of course!
        """
        # L = C1 * νc**3 / (exp (C2 νc / [αTb + β]) − 1)
        #
        # C1 = 2 * h * c**2 and C2 = hc/k
        #
        lut = kwargs.get('lut', None)
        normalized = kwargs.get('normalized', True)

        if lut is not None:
            raise NotImplementedError('Using a tb-radiance LUT is not yet supported')
        if not normalized:
            raise NotImplementedError('Deriving the band integrated radiance is not supported')

        c_1 = 2 * H_PLANCK * C_SPEED ** 2
        c_2 = H_PLANCK * C_SPEED / K_BOLTZMANN

        vc_ = SEVIRI[self.bandname][self.platform_name][0]
        # Multiply by 100 to get SI units!
        vc_ *= 100.0
        alpha = SEVIRI[self.bandname][self.platform_name][1]
        beta = SEVIRI[self.bandname][self.platform_name][2]

        radiance = c_1 * vc_ ** 3 / \
            (np.exp(c_2 * vc_ / (alpha * tb_ + beta)) - 1)

        unit = 'W/m^2 sr^-1 (m^-1)^-1'
        scale = 1.0
        return {'radiance': radiance,
                'unit': unit,
                'scale': scale}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013-2022 Pytroll Developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The tests package."""

import doctest
import os
import sys

from pyspectral import blackbody, near_infrared_reflectance, solar

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

TRAVIS = os.environ.get("TRAVIS", False)
APPVEYOR = os.environ.get("APPVEYOR", False)


def suite():
    """Perform the unit testing."""
    mysuite = unittest.TestSuite()
    if not TRAVIS and not APPVEYOR:
        # Test sphinx documentation pages:
        mysuite.addTests(doctest.DocFileSuite('../../doc/usage.rst'))
        mysuite.addTests(doctest.DocFileSuite('../../doc/rad_definitions.rst'))
        # mysuite.addTests(doctest.DocFileSuite('../../doc/seviri_example.rst'))
        mysuite.addTests(doctest.DocFileSuite('../../doc/37_reflectance.rst'))
        # Test the documentation strings
        mysuite.addTests(doctest.DocTestSuite(solar))
        mysuite.addTests(doctest.DocTestSuite(near_infrared_reflectance))
        mysuite.addTests(doctest.DocTestSuite(blackbody))

    return mysuite


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Derive NIR reflectances of a a band in the 3-4 micron window region.

Derive the Near-Infrared reflectance of a given band in the solar and
thermal range (usually the 3.7-3.9 micron band) using a thermal atmospheric
window channel (usually around 11-12 microns).
"""

import logging
import os
import tempfile

import numpy as np

try:
    from dask.array import asanyarray, isnan, logical_or, where
except ImportError:
    from numpy import where, logical_or, asanyarray, isnan

from pyspectral.config import get_config
from pyspectral.radiance_tb_conversion import RadTbConverter
from pyspectral.solar import SolarIrradianceSpectrum
from pyspectral.utils import BANDNAMES, WAVE_LENGTH, get_bandname_from_wavelength

LOG = logging.getLogger(__name__)

EPSILON = 0.005
TB_MIN = 150.
TB_MAX = 360.
TERMINATOR_LIMIT = 85.0


class Calculator(RadTbConverter):
    """
    A thermal near-infrared (~3.7 micron) band reflectance calculator.

    Given the relative spectral response of the NIR band, the solar zenith
    angle, and the brightness temperatures of the NIR and the Thermal bands,
    derive the solar reflectance for the NIR band removing the thermal
    (terrestrial) part. The in-band solar flux over the NIR band is
    optional. If not provided, it will be calculated here!

    The relfectance calculated is without units and should be between 0 and 1.
    """

    def __init__(self, platform_name, instrument, band,
                 detector='det-1', wavespace=WAVE_LENGTH,
                 solar_flux=None, sunz_threshold=TERMINATOR_LIMIT, masking_limit=TERMINATOR_LIMIT):
        """Initialize the Class instance."""
        super(Calculator, self).__init__(platform_name, instrument, band, detector=detector, wavespace=wavespace)

        from numbers import Number
        self.bandname = None
        self.bandwavelength = None

        if isinstance(band, str):
            self.bandname = BANDNAMES.get(self.instrument, BANDNAMES['generic']).get(band, band)
            self.bandwavelength = self.rsr[self.bandname][
                'det-1']['central_wavelength']
        elif isinstance(band, Number):
            self.bandwavelength = band
            self.bandname = get_bandname_from_wavelength(self.instrument, band, self.rsr)

        if self.bandwavelength > 3.95 or self.bandwavelength < 3.5:
            raise NotImplementedError('NIR reflectance is not supported outside ' +
                                      'the 3.5-3.95 micron interval')

        options = get_config()

        self.solar_flux = solar_flux
        if self.solar_flux is None:
            self._get_solarflux()

        # The sun-zenith angle limit in degrees defining how far towards the
        # terminator we try derive a
        self.detector = detector
        self.sunz_threshold = sunz_threshold
        self.masking_limit = masking_limit
        self._rad3x = None
        self._rad3x_t11 = None
        self._solar_radiance = None
        self._rad3x_correction = 1.0
        self._r3x = None
        self._e3x = None
        self.lutfile = None

        platform_sensor = platform_name + '-' + instrument
        tb2rad_dir = options.get('tb2rad_dir', tempfile.gettempdir())
        if platform_sensor in options and 'tb2rad_lut_filename' in options[platform_sensor]:
            if isinstance(options[platform_sensor]['tb2rad_lut_filename'], dict):
                for item in options[platform_sensor]['tb2rad_lut_filename']:
                    if item == self.bandname or item == self.bandname.lower():
                        self.lutfile = options[platform_sensor]['tb2rad_lut_filename'][item]
                        break
                if self.lutfile is None:
                    LOG.warning("Failed determine LUT filename from config: %s", str(
                        options[platform_sensor]['tb2rad_lut_filename']))
            else:
                self.lutfile = options[platform_sensor]['tb2rad_lut_filename']

            if self.lutfile and not self.lutfile.endswith('.npz'):
                self.lutfile = self.lutfile + '.npz'

            if self.lutfile and not os.path.exists(os.path.dirname(self.lutfile)):
                LOG.warning(
                    "Directory %s does not exist! Check config file", os.path.dirname(self.lutfile))
                self.lutfile = os.path.join(tb2rad_dir, os.path.basename(self.lutfile))

        if self.lutfile is None:
            LOG.info("No lut filename available in config file. "
                     "Will generate filename automatically")
            lutname = 'tb2rad_lut_{0}_{1}_{band}'.format(
                self.platform_name.lower(), self.instrument.lower(), band=self.bandname.lower())
            self.lutfile = os.path.join(tb2rad_dir, lutname + '.npz')

        LOG.info("lut filename: " + str(self.lutfile))
        if not os.path.exists(self.lutfile):
            self.make_tb2rad_lut(self.lutfile)
            self.lut = self.read_tb2rad_lut(self.lutfile)
            LOG.info("LUT file created")
        else:
            self.lut = self.read_tb2rad_lut(self.lutfile)
            LOG.info("File was there and has been read!")

    def derive_rad39_corr(self, bt11, bt13, method='rosenfeld'):
        """Derive the CO2 correction to be applied to the 3.9 channel.

        Derive the 3.9 radiance correction factor to account for the
        attenuation of the emitted 3.9 radiance by CO2
        absorption. Requires the 11 micron window band and the 13.4
        CO2 absorption band, as e.g. available on SEVIRI. Currently
        only supports the Rosenfeld method

        """
        if method != 'rosenfeld':
            raise AttributeError("Only CO2 correction for SEVIRI using "
                                 "the Rosenfeld equation is supported!")

        LOG.debug("Derive the 3.9 micron radiance CO2 correction coefficent")
        self._rad3x_correction = (bt11 - 0.25 * (bt11 - bt13)) ** 4 / bt11 ** 4

    def _get_solarflux(self):
        """Derive the in-band solar flux from rsr over the Near IR band (3.7 or 3.9 microns)."""
        solar_spectrum = \
            SolarIrradianceSpectrum(dlambda=0.0005,
                                    wavespace=self.wavespace)
        self.solar_flux = solar_spectrum.inband_solarflux(self.rsr[self.bandname])

    def emissive_part_3x(self, tb=True):
        """Get the emissive part of the 3.x band."""
        try:
            # Emissive part:
            self._e3x = self._rad3x_t11 * (1 - self._r3x)
            # Use the original channel data on the night side
            self._e3x = where(isnan(self._e3x), self._rad3x, self._e3x)
            # Unsure how much sense it makes to apply the co2 correction term here!?
            # FIXME!
            # self._e3x *= self._rad3x_correction

        except TypeError:
            LOG.warning(
                "Couldn't derive the emissive part \n" +
                "Please derive the relfectance prior to requesting the emissive part")

        if tb:
            return self.radiance2tb(self._e3x)
        else:
            return self._e3x

    def reflectance_from_tbs(self, sun_zenith, tb_near_ir, tb_thermal, **kwargs):
        """Derive reflectances from Tb's in the 3.x band.

        The relfectance calculated is without units and should be between 0 and 1.

        Inputs:

          sun_zenith: Sun zenith angle for every pixel - in degrees

          tb_near_ir: The 3.7 (or 3.9 or equivalent) IR Tb's at every pixel
                      (Kelvin)

          tb_thermal: The 10.8 (or 11 or 12 or equivalent) IR Tb's at every
                      pixel (Kelvin)

          tb_ir_co2: The 13.4 micron channel (or similar - co2 absorption band)
                     brightness temperatures at every pixel. If None, no CO2
                     absorption correction will be applied.

        """
        # Check for dask arrays
        if hasattr(tb_near_ir, 'compute') or hasattr(tb_thermal, 'compute'):
            compute = False
        else:
            compute = True
        if hasattr(tb_near_ir, 'mask') or hasattr(tb_thermal, 'mask'):
            is_masked = True
        else:
            is_masked = False

        tb_nir = get_as_array(tb_near_ir)
        tb_therm = get_as_array(tb_thermal)
        sun_zenith = get_as_array(sun_zenith)

        if tb_therm.shape != tb_nir.shape:
            errmsg = 'Dimensions do not match! {0} and {1}'.format(
                str(tb_therm.shape), str(tb_nir.shape))
            raise ValueError(errmsg)

        tb_ir_co2 = kwargs.get('tb_ir_co2')
        lut = kwargs.get('lut', self.lut)

        if tb_ir_co2 is None:
            co2corr = False
            tbco2 = None
        else:
            co2corr = True
            tbco2 = get_as_array(tb_ir_co2)

        if not self.rsr:
            raise NotImplementedError("Reflectance calculations without "
                                      "rsr not yet supported!")

        # Assume rsr is in microns!!!
        # FIXME!
        self._rad3x_t11 = self.tb2radiance(tb_therm, lut=lut)['radiance']
        rsr_integral = tb_therm.dtype.type(self.rsr_integral)
        thermal_emiss_one = self._rad3x_t11 * rsr_integral
        l_nir = self.tb2radiance(tb_nir, lut=lut)['radiance']
        self._rad3x = l_nir.copy()
        l_nir *= rsr_integral

        if thermal_emiss_one.ravel().shape[0] < 10:
            LOG.info('thermal_emiss_one = %s', str(thermal_emiss_one))
        if l_nir.ravel().shape[0] < 10:
            LOG.info('l_nir = %s', str(l_nir))

        LOG.debug("Apply sun-zenith angle clipping between 0 and %5.2f", self.masking_limit)

        sunz = sun_zenith.clip(0, self.sunz_threshold)
        mu0 = np.cos(np.deg2rad(sunz))

        # mu0 = np.where(np.less(mu0, 0.1), 0.1, mu0)
        self._solar_radiance = (self.solar_flux * mu0 / np.pi).astype(tb_nir.dtype)

        # CO2 correction to the 3.9 radiance, only if tbs of a co2 band around
        # 13.4 micron is provided:
        if co2corr:
            self.derive_rad39_corr(tb_therm, tbco2)
            LOG.info("CO2 correction applied...")
        else:
            self._rad3x_correction = np.float64(1.0)
        self._rad3x_correction = self._rad3x_correction.astype(tb_nir.dtype)

        corrected_thermal_emiss_one = thermal_emiss_one * self._rad3x_correction
        nomin = l_nir - corrected_thermal_emiss_one
        denom = self._solar_radiance - corrected_thermal_emiss_one
        data = nomin / denom
        mask = denom < EPSILON

        if self.masking_limit is not None:
            sunzmask = (sun_zenith < 0.0) | (sun_zenith > self.masking_limit)
            logical_or(sunzmask, mask, out=mask)
        logical_or(mask, np.isnan(tb_nir), out=mask)

        self._r3x = where(mask, np.nan, data)

        # Reflectances should be between 0 and 1, but values above 1 is
        # perfectly possible and okay! (Multiply by 100 to get reflectances
        # in percent)
        if hasattr(self._r3x, 'compute') and compute:
            res = self._r3x.compute()
        else:
            res = self._r3x
        if is_masked:
            res = np.ma.masked_invalid(res)
        return res


def get_as_array(variable):
    """Return variable as a Dask or Numpy array.

    Variable may be a scalar, a list or a Numpy/Dask array.
    """
    if np.isscalar(variable):
        return asanyarray([variable, ])

    return asanyarray(variable)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2024 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility functions."""

import logging
import os
import tarfile
import warnings
from functools import wraps
from inspect import getfullargspec

import numpy as np
import requests

from pyspectral.bandnames import BANDNAMES
from pyspectral.config import get_config

TQDM_LOADED = True
try:
    from tqdm import tqdm
except ImportError:
    TQDM_LOADED = False


LOG = logging.getLogger(__name__)

WAVE_LENGTH = 'wavelength'
WAVE_NUMBER = 'wavenumber'

INSTRUMENTS = {'Envisat': 'aatsr',
               'GOES-16': 'abi',
               'GOES-17': 'abi',
               'GOES-18': 'abi',
               'GOES-19': 'abi',
               'FY-4A': 'agri',
               'FY-4B': ['agri', 'ghi'],
               'Himawari-8': 'ahi',
               'Himawari-9': 'ahi',
               'GEO-KOMPSAT-2A': 'ami',
               'GEO-KOMPSAT-2B': 'goci-2',
               'NOAA-10': 'avhrr/1',
               'NOAA-6': 'avhrr/1',
               'NOAA-8': 'avhrr/1',
               'TIROS-N': 'avhrr/1',
               'NOAA-11': 'avhrr/2',
               'NOAA-12': 'avhrr/2',
               'NOAA-14': 'avhrr/2',
               'NOAA-7': 'avhrr/2',
               'NOAA-9': 'avhrr/2',
               'Metop-A': 'avhrr/3',
               'Metop-B': 'avhrr/3',
               'Metop-C': 'avhrr/3',
               'NOAA-15': 'avhrr/3',
               'NOAA-16': 'avhrr/3',
               'NOAA-17': 'avhrr/3',
               'NOAA-18': 'avhrr/3',
               'NOAA-19': 'avhrr/3',
               'HY-1C': 'cocts',
               'Meteosat-12': 'fci',
               'MTG-I1': 'fci',
               'FY-3D': 'mersi-2',
               'FY-3G': 'mersi-rm',
               'Metop-SG-A1': 'metimage',
               'EOS-Aqua': 'modis',
               'EOS-Terra': 'modis',
               'Sentinel-2A': 'msi',
               'Sentinel-2B': 'msi',
               'Arctica-M-N1': 'msu-gsa',
               'Electro-L-N2': 'msu-gs',
               'Sentinel-3A': ['olci', 'slstr'],
               'Sentinel-3B': ['olci', 'slstr'],
               'Landsat-8': 'oli',
               'Meteosat-10': 'seviri',
               'Meteosat-11': 'seviri',
               'Meteosat-8': 'seviri',
               'Meteosat-9': 'seviri',
               'NOAA-20': 'viirs',
               'NOAA-21': 'viirs',
               'Suomi-NPP': 'viirs',
               'FY-3A': ['virr', 'mersi-1'],
               'FY-3B': ['virr', 'mersi-1'],
               'FY-3C': ['virr', 'mersi-1'],
               'DSCOVR': 'epic'}


INSTRUMENT_TRANSLATION_DASH2SLASH = {'avhrr-1': 'avhrr/1',
                                     'avhrr-2': 'avhrr/2',
                                     'avhrr-3': 'avhrr/3'}

HTTP_PYSPECTRAL_RSR = "https://zenodo.org/records/11110033/files/pyspectral_rsr_data.tgz"

RSR_DATA_VERSION_FILENAME = "PYSPECTRAL_RSR_VERSION"
RSR_DATA_VERSION = "v1.3.0"

ATM_CORRECTION_LUT_VERSION = {}
ATM_CORRECTION_LUT_VERSION['antarctic_aerosol'] = {'version': 'v1.0.1',
                                                   'filename': 'PYSPECTRAL_ATM_CORR_LUT_AA'}
ATM_CORRECTION_LUT_VERSION['continental_average_aerosol'] = {'version': 'v1.0.1',
                                                             'filename': 'PYSPECTRAL_ATM_CORR_LUT_CAA'}
ATM_CORRECTION_LUT_VERSION['continental_clean_aerosol'] = {'version': 'v1.0.1',
                                                           'filename': 'PYSPECTRAL_ATM_CORR_LUT_CCA'}
ATM_CORRECTION_LUT_VERSION['continental_polluted_aerosol'] = {'version': 'v1.0.1',
                                                              'filename': 'PYSPECTRAL_ATM_CORR_LUT_CPA'}
ATM_CORRECTION_LUT_VERSION['desert_aerosol'] = {'version': 'v1.0.1',
                                                'filename': 'PYSPECTRAL_ATM_CORR_LUT_DA'}
ATM_CORRECTION_LUT_VERSION['marine_clean_aerosol'] = {'version': 'v1.0.1',
                                                      'filename': 'PYSPECTRAL_ATM_CORR_LUT_MCA'}
ATM_CORRECTION_LUT_VERSION['marine_polluted_aerosol'] = {'version': 'v1.0.1',
                                                         'filename': 'PYSPECTRAL_ATM_CORR_LUT_MPA'}
ATM_CORRECTION_LUT_VERSION['marine_tropical_aerosol'] = {'version': 'v1.0.1',
                                                         'filename': 'PYSPECTRAL_ATM_CORR_LUT_MTA'}
ATM_CORRECTION_LUT_VERSION['rural_aerosol'] = {'version': 'v1.0.1',
                                               'filename': 'PYSPECTRAL_ATM_CORR_LUT_RA'}
ATM_CORRECTION_LUT_VERSION['urban_aerosol'] = {'version': 'v1.0.1',
                                               'filename': 'PYSPECTRAL_ATM_CORR_LUT_UA'}
ATM_CORRECTION_LUT_VERSION['rayleigh_only'] = {'version': 'v1.0.1',
                                               'filename': 'PYSPECTRAL_ATM_CORR_LUT_RO'}

#: Aerosol types available as downloadable LUTs for rayleigh correction
AEROSOL_TYPES = ['antarctic_aerosol', 'continental_average_aerosol',
                 'continental_clean_aerosol', 'continental_polluted_aerosol',
                 'desert_aerosol', 'marine_clean_aerosol',
                 'marine_polluted_aerosol', 'marine_tropical_aerosol',
                 'rayleigh_only', 'rural_aerosol', 'urban_aerosol']

ATMOSPHERES = {'subarctic summer': 4, 'subarctic winter': 5,
               'midlatitude summer': 6, 'midlatitude winter': 7,
               'tropical': 8, 'us-standard': 9}

HTTPS_RAYLEIGH_LUTS = {}
LUT_URL_PREFIX = "https://zenodo.org/record/1288441/files/pyspectral_atm_correction_luts"
for atype in AEROSOL_TYPES:
    name = {'rayleigh_only': 'no_aerosol'}.get(atype, atype)
    url = "{prefix}_{name}.tgz".format(prefix=LUT_URL_PREFIX, name=name)
    HTTPS_RAYLEIGH_LUTS[atype] = url


def get_rayleigh_lut_dir(aerosol_type):
    """Get the rayleigh LUT directory for the specified aerosol type."""
    conf = get_config()
    local_rayleigh_dir = conf.get('rayleigh_dir')
    return os.path.join(local_rayleigh_dir, aerosol_type)


def convert2wavenumber(rsr):
    """Convert Spectral Responses from wavelength to wavenumber space.

    Take rsr data set with all channels and detectors for an instrument
    each with a set of wavelengths and normalised responses and
    convert to wavenumbers and responses

    :rsr: Relative Spectral Response function (all bands)
    Returns:
      :retv: Relative Spectral Responses in wave number space
      :info: Dictionary with scale (to go convert to SI units) and unit

    """
    retv = {}
    for chname in rsr.keys():  # Go through bands/channels
        retv[chname] = {}
        for det in rsr[chname].keys():  # Go through detectors
            retv[chname][det] = {}
            if 'wavenumber' in rsr[chname][det].keys():
                # Make a copy. Data are already in wave number space
                retv[chname][det] = rsr[chname][det].copy()
                LOG.debug("RSR data already in wavenumber space. No conversion needed.")
                continue

            for sat in rsr[chname][det].keys():
                if sat == "wavelength":
                    # micro meters to cm
                    wnum = 1. / (1e-4 * rsr[chname][det][sat])
                    retv[chname][det]['wavenumber'] = wnum[::-1]
                elif sat == "response":
                    # Flip the response array:
                    if isinstance(rsr[chname][det][sat], dict):
                        retv[chname][det][sat] = {}
                        for name in rsr[chname][det][sat].keys():
                            resp = rsr[chname][det][sat][name]
                            retv[chname][det][sat][name] = resp[::-1]
                    else:
                        resp = rsr[chname][det][sat]
                        retv[chname][det][sat] = resp[::-1]

    unit = 'cm-1'
    si_scale = 100.0
    return retv, {'unit': unit, 'si_scale': si_scale}


def get_central_wave(wav, resp, weight=1.0):
    """Calculate the central wavelength or the central wavenumber.

    Calculate the central wavelength or the central wavenumber, depending on
    which parameters is input.  On default the weighting funcion is
    f(lambda)=1.0, but it is possible to add a custom weight, e.g. f(lambda) =
    1./lambda**4 for Rayleigh scattering calculations

    """
    # info: {'unit': unit, 'si_scale': si_scale}
    # To get the wavelenght/wavenumber in SI units (m or m-1):
    # wav = wav * info['si_scale']

    # res = np.trapz(resp*wav, wav) / np.trapz(resp, wav)
    # Check if it is a wavelength or a wavenumber and convert to microns or cm-1:
    # This should perhaps be user defined!?
    # if info['unit'].find('-1') > 0:
    # Wavenumber:
    #     res *=
    return np.trapz(resp * wav * weight, wav) / np.trapz(resp * weight, wav)


def get_bandname_from_wavelength(sensor, wavelength, rsr, epsilon=0.1, multiple_bands=False):
    """Get the bandname from h5 rsr provided the approximate wavelength."""
    chdist_min = 2.0
    chfound = []
    for channel in rsr:
        chdist = abs(
            rsr[channel]['det-1']['central_wavelength'] - wavelength)
        if chdist < chdist_min and chdist < epsilon:
            chfound.append(BANDNAMES.get(sensor, BANDNAMES['generic']).get(channel, channel))

    if len(chfound) == 1:
        return chfound[0]
    if len(chfound) > 1:
        bstrlist = ['band={}'.format(b) for b in chfound]
        if not multiple_bands:
            raise AttributeError("More than one band found with that wavelength! {}".format(str(bstrlist)))
        LOG.debug("More than one band found with requested wavelength: %s", str(bstrlist))
        return chfound
    return None


def sort_data(x_vals, y_vals):
    """Sort the data so that x is monotonically increasing and contains no duplicates."""
    # Sort data
    # (This is needed in particular for EOS-Terra responses, as there are duplicates)
    idxs = np.argsort(x_vals)
    x_vals = x_vals[idxs]
    y_vals = y_vals[idxs]

    # De-duplicate data
    mask = np.r_[True, (np.diff(x_vals) > 0)]
    if not mask.all():
        numof_duplicates = np.repeat(mask, np.equal(mask, False)).shape[0]
        LOG.debug("Number of duplicates in the response function: %d - removing them",
                  numof_duplicates)
    x_vals = x_vals[mask]
    y_vals = y_vals[mask]

    return x_vals, y_vals


def convert2hdf5(ClassIn, platform_name, bandnames, scale=1e-06, detectors=None):
    """Retrieve original RSR data and convert to internal hdf5 format.

    *scale* is the number which has to be multiplied to the wavelength data in
    order to get it in the SI unit meter

    """
    import h5py

    instr = ClassIn(bandnames[0], platform_name)
    instr_name = instr.instrument.replace('/', '')
    filename = os.path.join(instr.output_dir,
                            "rsr_{0}_{1}.h5".format(instr_name,
                                                    platform_name))

    with h5py.File(filename, "w") as h5f:
        h5f.attrs['description'] = ('Relative Spectral Responses for ' +
                                    instr.instrument.upper())
        h5f.attrs['platform_name'] = platform_name
        h5f.attrs['band_names'] = bandnames

        for chname in bandnames:
            sensor = ClassIn(chname, platform_name)
            grp = h5f.create_group(chname)

            # If multiple detectors, assume all have same wavelength range in SRF.
            if detectors is not None:
                wvl = sensor.rsr[detectors[0]]['wavelength'][~np.isnan(sensor.rsr[detectors[0]]['wavelength'])]
                arr = sensor.rsr[detectors[0]]['wavelength']
                grp.attrs['number_of_detectors'] = len(detectors)
            else:
                wvl = sensor.rsr['wavelength'][~np.isnan(sensor.rsr['wavelength'])]
                arr = sensor.rsr['wavelength']

            # Save wavelengths to file
            dset = grp.create_dataset('wavelength', arr.shape, dtype='f')
            dset.attrs['unit'] = 'm'
            dset.attrs['scale'] = scale
            dset[...] = arr

            # Now to do the responses
            if detectors is None:
                rsp = sensor.rsr['response'][~np.isnan(sensor.rsr['wavelength'])]
                grp.attrs['central_wavelength'] = get_central_wave(wvl, rsp)
                arr = sensor.rsr['response']
                dset = grp.create_dataset('response', arr.shape, dtype='f')
                dset[...] = arr
            else:
                for cur_det in detectors:
                    det_grp = grp.create_group(cur_det)
                    rsp = sensor.rsr[cur_det]['response'][~np.isnan(sensor.rsr[cur_det]['wavelength'])]
                    det_grp.attrs['central_wavelength'] = get_central_wave(wvl, rsp)
                    arr = sensor.rsr[cur_det]['response']
                    dset = det_grp.create_dataset('response', arr.shape, dtype='f')
                    dset[...] = arr


def download_rsr(dest_dir=None, dry_run=False):
    """Download the relative spectral response functions.

    Download the pre-compiled HDF5 formatted relative spectral response
    functions from the internet as tarballs, extracts them, then deletes
    the tarball.

    See :func:`pyspectral.rsr_reader.check_and_download` for a "smart" version
    of this process that only downloads the necessary files.

    Args:
        dest_dir (str): Path to put the temporary tarball and extracted RSR
            files.
        dry_run (bool): If True, don't actually download files, only log what
            URLs would be downloaded. Defaults to False.

    """
    config = get_config()
    local_rsr_dir = config.get('rsr_dir')
    dest_dir = dest_dir or local_rsr_dir

    LOG.info("Download RSR files and store in directory %s", dest_dir)
    filename = os.path.join(dest_dir, "pyspectral_rsr_data.tgz")
    LOG.debug("RSR URL: %s", HTTP_PYSPECTRAL_RSR)
    LOG.debug("Destination = %s", dest_dir)
    if dry_run:
        return

    _download_tarball_and_extract(HTTP_PYSPECTRAL_RSR, filename, dest_dir)


def download_luts(aerosol_types=None, dry_run=False, aerosol_type=None):
    """Download the luts from internet.

    See :func:`pyspectral.rayleigh.check_and_download` for a "smart" version
    of this process that only downloads the necessary files.

    Args:
        aerosol_types (Iterable): Aerosol types to download the LUTs for.
            Defaults to all aerosol types. See :data:`AEROSOL_TYPES` for the
            full list.
        dry_run (bool): If True, don't actually download files, only log what
            URLs would be downloaded. Defaults to False.
        aerosol_type (str): Deprecated.

    """
    aerosol_types = _get_aerosol_types(aerosol_types, aerosol_type)
    for subname in aerosol_types:
        LOG.debug('Aerosol type: %s', subname)
        lut_tarball_url = HTTPS_RAYLEIGH_LUTS[subname]
        LOG.debug('Atmospheric LUT URL = %s', lut_tarball_url)

        subdir_path = get_rayleigh_lut_dir(subname)
        LOG.debug('Create directory: %s', subdir_path)
        if not dry_run:
            os.makedirs(subdir_path, exist_ok=True)
        if dry_run:
            continue

        local_tarball_pathname = os.path.join(subdir_path, "pyspectral_rayleigh_correction_luts.tgz")
        _download_tarball_and_extract(lut_tarball_url, local_tarball_pathname, subdir_path)


def _get_aerosol_types(aerosol_types, aerosol_type):
    if aerosol_type is not None:
        warnings.warn("'aerosol_type' is deprecated, use 'aerosol_types' instead.", UserWarning,
                      stacklevel=3)
        if isinstance(aerosol_type, (list, tuple, set)):
            aerosol_types = aerosol_type
        else:
            aerosol_types = [aerosol_type]
    elif aerosol_types is None:
        aerosol_types = list(HTTPS_RAYLEIGH_LUTS.keys())
    return aerosol_types


def _download_tarball_and_extract(tarball_url, local_pathname, extract_dir):
    chunk_size = 1024 * 1024  # 1 MB
    response = requests.get(tarball_url)
    total_size = int(response.headers['content-length'])

    with open(local_pathname, "wb") as handle:
        for data in _tqdm_or_iter(response.iter_content(chunk_size=chunk_size),
                                  total=(int(total_size / chunk_size + 0.5)),
                                  unit='kB'):
            handle.write(data)

    tar = tarfile.open(local_pathname)
    tar.extractall(extract_dir)
    tar.close()
    os.remove(local_pathname)


def _tqdm_or_iter(an_iterable, **tqdm_kwargs):
    """Wrap an iterable with tqdm if it is available, otherwise return the iterable."""
    if TQDM_LOADED:
        return tqdm(iterable=an_iterable, **tqdm_kwargs)
    else:
        return an_iterable


def debug_on():
    """Turn debugging logging on."""
    logging_on(logging.DEBUG)


_is_logging_on = False


def logging_on(level=logging.WARNING):
    """Turn logging on."""
    global _is_logging_on

    if not _is_logging_on:
        console = logging.StreamHandler()
        console.setFormatter(logging.Formatter("[%(levelname)s: %(asctime)s :"
                                               " %(name)s] %(message)s",
                                               '%Y-%m-%d %H:%M:%S'))
        console.setLevel(level)
        logging.getLogger('').addHandler(console)
        _is_logging_on = True

    log = logging.getLogger('')
    log.setLevel(level)
    for h in log.handlers:
        h.setLevel(level)


class NullHandler(logging.Handler):
    """Empty handler."""

    def emit(self, record):
        """Record a message."""


def logging_off():
    """Turn logging off."""
    logging.getLogger('').handlers = [NullHandler()]


def get_logger(name):
    """Return logger with null handle."""
    log = logging.getLogger(name)
    if not log.handlers:
        log.addHandler(NullHandler())
    return log


def get_wave_range(in_chan, threshold=0.15):
    """Return central, min and max wavelength in an RSR greater than threshold.

    An RSR function will generally start near zero, increase to a maximum and
    then drop back to near zero. This function takes advantage of this to find
    the first and last points where the RSR is greater than a threshold. These
    points are then defined as the minimum and maximum wavelengths for a
    given channel, and can be used, for example, in Satpy reader YAML files.

    """
    cwl = get_central_wave(in_chan['wavelength'], in_chan['response'])

    wvls = in_chan['wavelength']
    rsr = in_chan['response']

    pts = (rsr > threshold).nonzero()
    min_wvl = wvls[pts[0][0]]
    max_wvl = wvls[pts[0][-1]]

    return [min_wvl, cwl, max_wvl]


def convert2str(value):
    """Convert a value to string.

    Args:
        value: Either a str, bytes or 1-element numpy array

    """
    value = bytes2string(value)
    return np2str(value)


def np2str(value):
    """Convert an ``numpy.bytes_`` to str.

    Note: ``numpy.string_`` was deprecated in numpy 2.0 in favor of
    ``numpy.bytes_``.

    Args:
        value (ndarray): scalar or 1-element numpy array to convert
    Raises:
        ValueError: if value is array larger than 1-element or it is not of
                    type `numpy.bytes_` or it is not a numpy array

    """
    if isinstance(value, str):
        return value

    if hasattr(value, 'dtype') and \
            issubclass(value.dtype.type, (np.str_, np.bytes_, np.object_)) \
            and value.size == 1:
        value = value.item()
        # python 3 - was scalar numpy array of bytes
        # otherwise python 2 - scalar numpy array of 'str'
        if not isinstance(value, str):
            return value.decode()
        return value

    raise ValueError("Array is not a string type or is larger than 1")


def bytes2string(var):
    """Decode a bytes variable and return a string."""
    if isinstance(var, bytes):
        return var.decode('utf-8')
    return var


def _replace_inst_name(instruments):
    """Replace an instrument name if it's not available on a platform."""
    if isinstance(instruments, list):
        return instruments[0]
    else:
        return instruments


def check_and_adjust_instrument_name(platform_name, instrument):
    """Check instrument name and try fix if inconsistent.

    It checks against the possible listed instrument names for each platform.
    It also makes an adjustment replacing names like avhrr/1 with avhrr1,
    removing the '/'.
    """
    instr = INSTRUMENTS.get(platform_name, instrument.lower())
    if isinstance(instr, list):
        for inst in instr:
            goodinst = are_instruments_identical(inst, instrument.lower())
            if goodinst:
                break
    else:
        goodinst = are_instruments_identical(instr, instrument.lower())

    if not goodinst:
        instrument = _replace_inst_name(instr)
        LOG.warning("Inconsistent instrument/satellite input - instrument set to %s",
                    instrument)

    return instrument.lower().replace('/', '').replace('-', '')


def are_instruments_identical(name1, name2):
    """Given two instrument names check if they are both describing the same instrument.

    Takes care of the case of AVHRR where the internal pyspectral naming
    (following WMO Oscar) is is with a slash as in 'avhrr/1', but where a
    naming using a dash instead is equally accepted, as in 'avhrr-1'.
    """
    if not isinstance(name1, str) or not isinstance(name2, str):
        raise ValueError("Instrument names must be strings.")

    if name1 == name2:
        return True
    if name1 == INSTRUMENT_TRANSLATION_DASH2SLASH.get(name2):
        return True
    return False


def use_map_blocks_on(argument_to_run_map_blocks_on):
    """Use map blocks on a given argument.

    This decorator assumes only one of the arguments of the decorated function is chunked.
    """
    def decorator(f):
        argspec = getfullargspec(f)
        argument_index = argspec.args.index(argument_to_run_map_blocks_on)

        @wraps(f)
        def wrapper(*args, **kwargs):
            array = args[argument_index]
            chunks = getattr(array, "chunks", None)
            if chunks is None:
                return f(*args, **kwargs)
            import dask.array as da
            import xarray as xr
            if isinstance(array, da.Array):
                return da.map_blocks(f, *args, **kwargs)
            elif isinstance(array, xr.DataArray):
                new_array = array.copy()
                new_args = list(args)
                new_args[argument_index] = array.data
                new_data = da.map_blocks(f, *new_args, **kwargs)
                new_array.data = new_data
                return new_array
            else:
                raise NotImplementedError(f"Don't know how to map_blocks on {type(array)}")
        return wrapper
    return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the 3.7 micron reflectance calculations."""

import unittest
from unittest.mock import patch

import numpy as np

from pyspectral.near_infrared_reflectance import TERMINATOR_LIMIT, Calculator, get_as_array

TEST_RSR = {'20': {},
            '99': {}}
TEST_RSR['20']['det-1'] = {}
TEST_RSR['20']['det-1']['central_wavelength'] = 3.780281935
TEST_RSR['20']['det-1']['wavelength'] = np.array([
    3.6123999, 3.6163599, 3.6264927, 3.6363862, 3.646468,
    3.6564937, 3.6664478, 3.6765388, 3.6865413, 3.6964585,
    3.7065142, 3.716509, 3.7264658, 3.7364102, 3.7463682,
    3.7563652, 3.7664226, 3.7763396, 3.7863384, 3.7964207,
    3.8063589, 3.8163606, 3.8264089, 3.8364836, 3.8463381,
    3.8563975, 3.8664163, 3.8763755, 3.8864797, 3.8964978,
    3.9064275, 3.9164873, 3.9264729, 3.9364026, 3.9465107,
    3.9535347], dtype='double')

TEST_RSR['20']['det-1']['response'] = np.array([
    0.01, 0.0118, 0.01987, 0.03226, 0.05028, 0.0849,
    0.16645, 0.33792, 0.59106, 0.81815, 0.96077, 0.92855,
    0.86008, 0.8661, 0.87697, 0.85412, 0.88922, 0.9541,
    0.95687, 0.91037, 0.91058, 0.94256, 0.94719, 0.94808,
    1., 0.92676, 0.67429, 0.44715, 0.27762, 0.14852,
    0.07141, 0.04151, 0.02925, 0.02085, 0.01414, 0.01], dtype='double')

TEST_RSR['99']['det-1'] = {}
TEST_RSR['99']['det-1']['central_wavelength'] = 10.8
TEST_RSR['99']['det-1']['wavelength'] = np.array([
    10.6123999,
    10.7563652,
    10.8563975,
    10.9064275,
    10.9535347], dtype='double')

TEST_RSR['99']['det-1']['response'] = np.array([
    0.33792,
    0.86008,
    1.,
    0.67429,
    0.07141], dtype='double')

TEST_RSR_WN = {'20': {}}
TEST_RSR_WN['20']['det-1'] = {}
WVN = np.array([2529.38208008, 2533.8840332, 2540.390625, 2546.81494141,
                2553.30883789, 2559.88354492, 2566.40722656, 2573.02270508,
                2579.72949219, 2586.37451172, 2593.09375, 2599.87548828,
                2606.55371094, 2613.41674805, 2620.29760742, 2627.18286133,
                2634.06005859, 2641.07421875, 2648.06713867, 2655.03930664,
                2662.14794922, 2669.25170898, 2676.36572266, 2683.50805664,
                2690.69702148, 2697.95288086, 2705.29199219, 2712.56982422,
                2719.94970703, 2727.43554688, 2734.8605957, 2742.38012695,
                2749.9831543, 2757.48510742, 2765.21142578, 2768.24291992],
               dtype='float32')
RESP = np.array([0.01, 0.01414, 0.02085, 0.02925, 0.04151,
                 0.07141, 0.14851999, 0.27761999, 0.44714999, 0.67429,
                 0.92676002, 1., 0.94808, 0.94718999, 0.94256002,
                 0.91057998, 0.91036999, 0.95687002, 0.95410001, 0.88922,
                 0.85412002, 0.87696999, 0.86610001, 0.86008, 0.92855,
                 0.96077001, 0.81814998, 0.59105998, 0.33792001, 0.16644999,
                 0.0849, 0.05028, 0.03226, 0.01987, 0.0118,
                 0.01], dtype='float32')

TEST_RSR_WN['20']['det-1']['central_wavelength'] = 3.780281935
TEST_RSR_WN['20']['det-1']['wavenumber'] = WVN
TEST_RSR_WN['20']['det-1']['response'] = RESP


class TestReflectance(unittest.TestCase):
    """Unit testing the reflectance calculations."""

    def test_rsr_integral(self):
        """Test calculating the integral of the relative spectral response function."""
        with patch('pyspectral.radiance_tb_conversion.RelativeSpectralResponse') as mymock:
            instance = mymock.return_value
            instance.rsr = TEST_RSR
            instance.unit = '1e-6 m'
            instance.si_scale = 1e-6

            refl37 = Calculator('EOS-Aqua', 'modis', '20')

        expected = 1.8563451e-07  # unit = 'm' (meter)
        np.testing.assert_allclose(refl37.rsr_integral, expected)

        with patch('pyspectral.radiance_tb_conversion.RelativeSpectralResponse') as mymock:
            instance = mymock.return_value
            instance.rsr = TEST_RSR_WN
            instance.unit = 'cm-1'
            instance.si_scale = 100.

            refl37 = Calculator('EOS-Aqua', 'modis', '20', wavespace='wavenumber')

        expected = 13000.385  # SI units = 'm-1' (1/meter)
        res = refl37.rsr_integral
        np.testing.assert_allclose(res / expected, 1.0, 6)

    def test_reflectance(self):
        """Test the derivation of the reflective part of a 3.7 micron band."""
        with patch('pyspectral.radiance_tb_conversion.RelativeSpectralResponse') as mymock:
            instance = mymock.return_value
            # VIIRS doesn't have a channel '20' like MODIS so the generic
            # mapping this test will end up using will find 'ch20' for VIIRS
            viirs_rsr = {'ch20': TEST_RSR['20'], '99': TEST_RSR['99']}
            instance.rsr = viirs_rsr
            instance.unit = '1e-6 m'
            instance.si_scale = 1e-6

            with self.assertRaises(NotImplementedError):
                dummy = Calculator('Suomi-NPP', 'viirs', 10.8)
                del dummy

            refl37 = Calculator('Suomi-NPP', 'viirs', 3.7)
            self.assertEqual(refl37.bandwavelength, 3.7)
            self.assertEqual(refl37.bandname, 'ch20')
            # Default sunz-threshold used to stay on day side and away from terminator:
            self.assertEqual(refl37.sunz_threshold, 85.0)

        with patch('pyspectral.radiance_tb_conversion.RelativeSpectralResponse') as mymock:
            instance = mymock.return_value
            instance.rsr = TEST_RSR
            instance.unit = '1e-6 m'
            instance.si_scale = 1e-6

            refl37 = Calculator('EOS-Aqua', 'modis', '20')
            self.assertEqual(refl37.sunz_threshold, TERMINATOR_LIMIT)
            self.assertEqual(refl37.masking_limit, TERMINATOR_LIMIT)

            refl37_sz88 = Calculator('EOS-Aqua', 'modis', '20', sunz_threshold=88.0, masking_limit=None)
            self.assertEqual(refl37_sz88.sunz_threshold, 88.0)
            self.assertIsNone(refl37_sz88.masking_limit)
            self.assertAlmostEqual(refl37_sz88.bandwavelength, 3.780282, 5)
            self.assertEqual(refl37_sz88.bandname, '20')

        sunz = 80.
        tb3 = 290.
        tb4 = 282.
        refl = refl37.reflectance_from_tbs(sunz, tb3, tb4)
        np.testing.assert_allclose(refl[0], 0.251245010648, 6)

        sunz = 85.
        tb3 = 290.
        tb4 = 282.
        refl = refl37.reflectance_from_tbs(sunz, tb3, tb4)
        np.testing.assert_allclose(refl[0], 1.12236884, 6)

        sunz = np.array([85.1])
        refl = refl37.reflectance_from_tbs(sunz, tb3, tb4)
        self.assertTrue(np.isnan(refl[0]))

        refl_sz88 = refl37_sz88.reflectance_from_tbs(sunz, tb3, tb4)
        np.testing.assert_allclose(refl_sz88[0], 1.2064644, 6)
        sunz = np.array([86.0])
        self.assertTrue(np.isnan(refl[0]))

        tb3x = refl37.emissive_part_3x()
        np.testing.assert_allclose(tb3x, 276.213054, 6)

        sunz = np.array([80.])
        tb3 = np.array([295.])
        tb4 = np.array([282.])
        refl = refl37.reflectance_from_tbs(sunz, tb3, tb4)
        np.testing.assert_allclose(refl[0], 0.452497961, 6)

        tb3x = refl37.emissive_part_3x()
        np.testing.assert_allclose(tb3x, 270.077268, 6)

        sunz = np.array([50.])
        tb3 = np.array([300.])
        tb4 = np.array([285.])
        refl = refl37.reflectance_from_tbs(sunz, tb3, tb4)
        np.testing.assert_allclose(refl[0], 0.1189217, 6)

        tb3x = refl37.emissive_part_3x()
        np.testing.assert_allclose(tb3x, 282.455426, 6)

        sunz = np.array([50.])
        tb3 = np.ma.masked_array([300.], mask=False)
        tb4 = np.ma.masked_array([285.], mask=False)
        refl = refl37.reflectance_from_tbs(sunz, tb3, tb4)
        self.assertTrue(hasattr(refl, 'mask'))

        sunz = np.array([80.], dtype=np.float32)
        tb3 = np.array([295.], dtype=np.float32)
        tb4 = np.array([282.], dtype=np.float32)
        refl = refl37.reflectance_from_tbs(sunz, tb3, tb4)
        np.testing.assert_allclose(refl, np.array([0.45249779], np.float32))
        assert refl.dtype == np.float32

        try:
            import dask.array as da
            import dask.config

            from pyspectral.tests.unittest_helpers import ComputeCountingScheduler

            with dask.config.set(scheduler=ComputeCountingScheduler(max_computes=0)):
                sunz = da.from_array([50.], chunks=10)
                tb3 = da.from_array([300.], chunks=10)
                tb4 = da.from_array([285.], chunks=10)
                refl = refl37.reflectance_from_tbs(sunz, tb3, tb4)
                self.assertTrue(hasattr(refl, 'compute'))
        except ImportError:
            pass


def test_get_as_array_from_scalar_input_dask():
    """Test the function to return an array when input is a scalar - using Dask."""
    res = get_as_array(2.3)
    if hasattr(res, 'compute'):
        assert res.compute()[0] == 2.3
    else:
        assert res[0] == 2.3


def test_get_as_array_from_scalar_input_numpy():
    """Test the function to return an array when input is a scalar - using Numpy."""
    with patch('pyspectral.near_infrared_reflectance.asanyarray', new=np.asanyarray):
        res = get_as_array(2.3)

    assert res[0] == 2.3


def test_get_as_array_from_numpy_array_input_dask():
    """Test the function to return an array when input is a numpy array - using Dask."""
    res = get_as_array(np.array([1.0, 2.0]))
    if hasattr(res, 'compute'):
        np.testing.assert_allclose(res.compute(), np.array([1.0, 2.0]), 5)
    else:
        np.testing.assert_allclose(res, np.array([1.0, 2.0]), 5)


def test_get_as_array_from_numpy_array_input_numpy():
    """Test the function to return an array when input is a numpy array - using Numpy."""
    from pyspectral.near_infrared_reflectance import get_as_array

    with patch('pyspectral.near_infrared_reflectance.asanyarray', new=np.asanyarray):
        res = get_as_array(np.array([1.0, 2.0]))

    np.testing.assert_allclose(res, np.array([1.0, 2.0]), 5)


def test_get_as_array_from_list_input_dask():
    """Test the function to return an array when input is a list - using Dask."""
    res = get_as_array([1.0, 2.0])
    if hasattr(res, 'compute'):
        np.testing.assert_allclose(res.compute(), np.array([1.0, 2.0]), 5)
    else:
        np.testing.assert_allclose(res, np.array([1.0, 2.0]), 5)


def test_get_as_array_from_list_input_numpy():
    """Test the function to return an array when input is a list - using Numpy."""
    with patch('pyspectral.near_infrared_reflectance.asanyarray', new=np.asanyarray):
        res = get_as_array([1.1, 2.2])

    np.testing.assert_allclose(res, np.array([1.1, 2.2]), 5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Atmospheric correction of shortwave imager bands in the wavelength range 400 to 800 nm."""

import logging
import numbers
import os

import h5py
import numpy as np

try:
    import dask.array as da
except ImportError:
    da = None

from geotiepoints.multilinear import MultilinearInterpolator

from pyspectral.config import get_config
from pyspectral.rsr_reader import RelativeSpectralResponse
from pyspectral.utils import (
    AEROSOL_TYPES,
    ATM_CORRECTION_LUT_VERSION,
    ATMOSPHERES,
    BANDNAMES,
    INSTRUMENTS,
    download_luts,
    get_central_wave,
    get_rayleigh_lut_dir,
)

LOG = logging.getLogger(__name__)


def _map_blocks_or_direct_call(func, *args, **kwargs):
    """Call dask's map_blocks or call func directly if dask is not available."""
    if da is None:
        kwargs.pop("meta", None)
        kwargs.pop("dtype", None)
        kwargs.pop("chunks", None)
        return func(*args, **kwargs)
    return da.map_blocks(func, *args, **kwargs)


def _clip_angles_inside_coordinate_range(zenith_angle, zenith_secant_max):
    """Clipping solar- or satellite zenith angles to be inside the allowed coordinate range.

    The "allowed" coordinate range is 0 to the angle corresponiding to a
    maximum secant. The maximum secant is here the maximum available in the
    Rayleigh LUT file.

    Also the nan's in the angle array are filled with zeros (0)!

    """
    clip_angle = np.nan_to_num(np.rad2deg(np.arccos(1. / zenith_secant_max)))
    zang = np.nan_to_num(zenith_angle)
    return np.clip(zang, 0, clip_angle)


class RayleighConfigBaseClass(object):
    """A base class for the Atmospheric correction, handling the configuration and LUT download."""

    def __init__(self, aerosol_type, atm_type='us-standard'):
        """Initialize class and determine LUT to use."""
        options = get_config()
        self.do_download = 'download_from_internet' in options and options['download_from_internet']
        self._lutfiles_version_uptodate = False

        if atm_type not in ATMOSPHERES:
            raise AttributeError('Atmosphere type not supported! ' +
                                 'Need to be one of {}'.format(str(ATMOSPHERES)))

        self._aerosol_type = aerosol_type
        if aerosol_type not in AEROSOL_TYPES:
            raise AttributeError('Aerosol type not supported! ' +
                                 'Need to be one of {0}'.format(str(AEROSOL_TYPES)))

        if atm_type not in ATMOSPHERES.keys():
            LOG.error("Atmosphere type %s not supported", atm_type)

        LOG.info("Atmosphere chosen: %s", atm_type)

        if (self._aerosol_type in ATM_CORRECTION_LUT_VERSION and
                self._get_lutfiles_version() == ATM_CORRECTION_LUT_VERSION[self._aerosol_type]['version']):
            self.lutfiles_version_uptodate = True

    def _get_lutfiles_version(self):
        """Get LUT file version.

        Check the version of the atm correction luts from the version file in the
        specific aerosol correction directory.

        """
        basedir = get_rayleigh_lut_dir(self._aerosol_type)
        lutfiles_version_path = os.path.join(basedir,
                                             ATM_CORRECTION_LUT_VERSION[self._aerosol_type]['filename'])

        if not os.path.exists(lutfiles_version_path):
            return "v0.0.0"

        with open(lutfiles_version_path, 'r') as fpt:
            # Get the version from the file
            return fpt.readline().strip()

    @property
    def lutfiles_version_uptodate(self):
        """Tell whether LUT file is up to date or not."""
        return self._lutfiles_version_uptodate

    @lutfiles_version_uptodate.setter
    def lutfiles_version_uptodate(self, value):
        """Store whether LUT file is up to date or not."""
        self._lutfiles_version_uptodate = value


class Rayleigh(RayleighConfigBaseClass):
    """
    Container for the atmospheric correction of satellite imager bands.

    This class removes background contributions of Rayleigh scattering of
    molecules and Mie scattering and absorption by aerosols.

    """

    def __init__(self, platform_name, sensor, **kwargs):
        """Initialize class and determine LUT to use."""
        atm_type = kwargs.get('atmosphere', 'us-standard')
        aerosol_type = kwargs.get('aerosol_type', 'marine_clean_aerosol')

        super(Rayleigh, self).__init__(aerosol_type, atm_type)

        self.platform_name = platform_name
        self.sensor = sensor
        self.coeff_filename = None

        # Try to fix instrument naming
        instr = INSTRUMENTS.get(platform_name, sensor)
        if instr != sensor:
            if isinstance(instr, list):
                if sensor not in instr:
                    raise ValueError("This satellite has multiple sensors, you must explicitly state which to use.")
            else:
                sensor = instr
                LOG.warning("Inconsistent sensor/satellite input - " +
                            "sensor set to %s", sensor)

        self.sensor = sensor.replace('/', '')

        rayleigh_dir = get_rayleigh_lut_dir(aerosol_type)
        ext = atm_type.replace(' ', '_')
        lutname = "rayleigh_lut_{0}.h5".format(ext)
        self.reflectance_lut_filename = os.path.join(rayleigh_dir, lutname)

        if not self._lutfiles_version_uptodate and self.do_download:
            LOG.info("Will download from internet...")
            download_luts(aerosol_type=aerosol_type)

        if (not os.path.exists(self.reflectance_lut_filename) or
                not os.path.isfile(self.reflectance_lut_filename)):
            raise IOError('pyspectral file for Rayleigh scattering correction ' +
                          'does not exist! Filename = ' +
                          str(self.reflectance_lut_filename))

        LOG.debug('LUT filename: %s', str(self.reflectance_lut_filename))

    def _get_effective_wavelength_and_band_name(self, band_name_or_wavelength):
        """Get the effective wavelength in nanometers and name of the band/channel.

        The provided argument can be either the name of a band/channel that
        pyspectral knows about in its RSR files or it can be a wavelength
        specified in micrometers. If a wavelength is provided it is converted
        to nanometers and returned. If a band name is provided then it is used
        to lookup the spectral response and determine an effective wavelength
        useful for rayleigh correction.

        Returns:
            A two element tuple of (wavelength in nanometers, name of band). If
            a wavelength is provided then the band name is a string
            representation of the wavelength.

        """
        # Get wavelength in nm for band:
        if isinstance(band_name_or_wavelength, numbers.Real):
            LOG.warning('A wavelength is provided instead of band name - ' +
                        'disregard the relative spectral responses and assume ' +
                        'it is the effective wavelength: %f (micro meter)', band_name_or_wavelength)
            wvl = band_name_or_wavelength
            band_name = f'{wvl:f}um'
        else:
            band_name = band_name_or_wavelength
            wvl = _get_rsr_wavelength_from_band_name(
                self.platform_name,
                self.sensor,
                band_name,
            )
            band_name_map = BANDNAMES.get(self.sensor, BANDNAMES['generic'])
            band_name_or_wavelength = band_name_map.get(band_name, band_name)
        LOG.debug("Band name: %s  Effective wavelength: %fum", band_name_or_wavelength, wvl)
        return wvl * 1000.0, band_name

    @staticmethod
    def _interp_rayleigh_refl_by_angles(sun_zenith, sat_zenith, azidiff,
                                        rayleigh_refl, reflectance_lut_filename):
        azid_coord, satz_sec_coord, sunz_sec_coord = get_reflectance_lut_from_file(
            reflectance_lut_filename)
        azid_coord = azid_coord.astype(rayleigh_refl.dtype, copy=False)
        satz_sec_coord = satz_sec_coord.astype(rayleigh_refl.dtype, copy=False)
        sunz_sec_coord = sunz_sec_coord.astype(rayleigh_refl.dtype, copy=False)

        sun_zenith = _clip_angles_inside_coordinate_range(sun_zenith, sunz_sec_coord.max())
        sunzsec = 1. / np.cos(np.deg2rad(sun_zenith))

        sat_zenith = _clip_angles_inside_coordinate_range(sat_zenith, satz_sec_coord.max())
        satzsec = 1. / np.cos(np.deg2rad(sat_zenith))

        smin = [sunz_sec_coord[0], azid_coord[0], satz_sec_coord[0]]
        smax = [sunz_sec_coord[-1], azid_coord[-1], satz_sec_coord[-1]]
        orders = [len(sunz_sec_coord), len(azid_coord), len(satz_sec_coord)]
        f_3d_grid = np.atleast_2d(rayleigh_refl.ravel())

        minterp = MultilinearInterpolator(smin, smax, orders, dtype=rayleigh_refl.dtype)
        minterp.set_values(f_3d_grid)
        interp_points2 = np.vstack((sunzsec.ravel(), 180 - azidiff.ravel(), satzsec.ravel()))
        res = minterp(interp_points2)
        res *= 100
        return res.reshape(sunzsec.shape)

    def get_reflectance(self, sun_zenith, sat_zenith, azidiff,
                        band_name_or_wavelength, redband=None):
        """Get the reflectance from the three sun-sat angles."""
        # if the user gave us non-dask arrays we'll use the dask
        # version of the algorithm but return numpy arrays back
        compute = da is not None and not isinstance(sun_zenith, da.Array)

        wvl, band_name = self._get_effective_wavelength_and_band_name(band_name_or_wavelength)
        repr_arr = sun_zenith if redband is None else redband
        try:
            rayleigh_refl = _get_wavelength_adjusted_lut_rayleigh_reflectance(
                self.reflectance_lut_filename, wvl)
        except ValueError:
            LOG.warning("Effective wavelength for band %s outside "
                        "nominal 400-800 nm range!", str(band_name))
            LOG.info("Setting the rayleigh/aerosol reflectance contribution to zero!")
            zeros_like = np.zeros_like if isinstance(repr_arr, np.ndarray) else da.zeros_like
            res = zeros_like(repr_arr)
        else:
            rayleigh_refl = rayleigh_refl.astype(repr_arr.dtype, copy=False)
            res = _map_blocks_or_direct_call(self._interp_rayleigh_refl_by_angles,
                                             sun_zenith, sat_zenith, azidiff, rayleigh_refl,
                                             self.reflectance_lut_filename,
                                             meta=np.array((), dtype=repr_arr.dtype),
                                             dtype=repr_arr.dtype,
                                             chunks=getattr(azidiff, "chunks", None))

        if redband is not None:
            res = _map_blocks_or_direct_call(self._relax_rayleigh_refl_correction_where_cloudy,
                                             redband, res,
                                             meta=np.array((), dtype=res.dtype),
                                             dtype=res.dtype,
                                             chunks=getattr(res, "chunks", None))

        res = np.clip(res, 0, 100)
        if compute:
            res = res.compute()
        return res

    @staticmethod
    def _relax_rayleigh_refl_correction_where_cloudy(redband, rayleigh_refl):
        return np.where(redband < 20., rayleigh_refl,
                        (1 - (redband - 20) / 80) * rayleigh_refl)

    @staticmethod
    def reduce_rayleigh_highzenith(zenith, rayref, thresh_zen, maxzen, strength):
        """Reduce the Rayleigh correction amount at high zenith angles.

        This linearly scales the Rayleigh reflectance, `rayref`, for solar or satellite zenith angles, `zenith`,
        above a threshold angle, `thresh_zen`. Between `thresh_zen` and `maxzen` the Rayleigh reflectance will
        be linearly scaled, from one at `thresh_zen` to zero at `maxzen`.
        """
        LOG.info("Reducing Rayleigh effect at high zenith angles.")
        factor = 1. - strength * np.where(zenith < thresh_zen, 0, (zenith - thresh_zen) / (maxzen - thresh_zen))
        # For low zenith factor can be greater than one, so we need to clip it into a sensible range.
        factor = np.clip(factor, 0, 1)
        return rayref * factor


def _get_rsr_wavelength_from_band_name(platform_name, sensor, band_name):
    try:
        rsr = RelativeSpectralResponse(platform_name, sensor)
    except OSError:
        LOG.exception(
            "No spectral responses for this platform and sensor: %s %s", platform_name, sensor)
        msg = ("Can't get effective wavelength for band %s on platform %s and sensor %s" %
               (str(band_name), platform_name, sensor))
        raise KeyError(msg)

    wvl = rsr.rsr[band_name]['det-1']['wavelength']
    resp = rsr.rsr[band_name]['det-1']['response']
    cwvl = get_central_wave(wvl, resp, weight=1. / wvl**4)
    return cwvl


def _get_wavelength_adjusted_lut_rayleigh_reflectance(lut_filename, wvl):
    with h5py.File(lut_filename, 'r') as h5f:
        rayleigh_refl = h5f["reflectance"]
        wvl_coord = h5f["wavelengths"][:]
        if not wvl_coord.min() < wvl < wvl_coord.max():
            raise ValueError("Wavelength out of range for available LUT wavelengths")
        wavelength_index, wavelength_factor = _get_wavelength_index_and_factor(wvl_coord, wvl)
        raylwvl = (wavelength_factor * rayleigh_refl[wavelength_index - 1, :, :, :] +
                   (1 - wavelength_factor) * rayleigh_refl[wavelength_index, :, :, :])
    return raylwvl


def _get_wavelength_index_and_factor(wvl_coord, wvl):
    wavelength_index = np.searchsorted(wvl_coord, wvl)
    wvl1 = wvl_coord[wavelength_index - 1]
    wvl2 = wvl_coord[wavelength_index]
    wavelength_factor = (wvl2 - wvl) / (wvl2 - wvl1)
    return wavelength_index, wavelength_factor


def get_reflectance_lut_from_file(lut_filename):
    """Get reflectance LUT.

    Read the Look-Up Tables from file with reflectances as a function of
    wavelength, satellite zenith secant, azimuth difference angle, and sun
    zenith secant.

    """
    with h5py.File(lut_filename, 'r') as h5f:
        azidiff = h5f['azimuth_difference']
        satellite_zenith_secant = h5f['satellite_zenith_secant']
        sun_zenith_secant = h5f['sun_zenith_secant']

        # load all of the data we are going to use in to memory
        azidiff = azidiff[:]
        satellite_zenith_secant = satellite_zenith_secant[:]
        sun_zenith_secant = sun_zenith_secant[:]

    return azidiff, satellite_zenith_secant, sun_zenith_secant


def check_and_download(dry_run=False, aerosol_types=None):
    """Download atm correction LUT tables if they are not up-to-date already.

    Do a check for the version of the atmospheric correction LUTs and attempt
    downloading only if needed.

    """
    aerosol_types = aerosol_types or AEROSOL_TYPES
    needed_aerosol_types = []
    for aerosol_type in aerosol_types:
        atmcorr = RayleighConfigBaseClass(aerosol_type)
        if atmcorr.lutfiles_version_uptodate:
            LOG.info("Atm correction LUTs, for aerosol distribution %s, already the latest!",
                     aerosol_type)
        else:
            needed_aerosol_types.append(aerosol_type)

    # Download
    if needed_aerosol_types:
        download_luts(aerosol_types=needed_aerosol_types, dry_run=dry_run)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Reading the spectral responses in the internal pyspectral hdf5 format."""

import logging
import os
from glob import glob
from os.path import expanduser

import numpy as np

from pyspectral.bandnames import BANDNAMES
from pyspectral.config import get_config
from pyspectral.utils import (
    INSTRUMENTS,
    RSR_DATA_VERSION,
    RSR_DATA_VERSION_FILENAME,
    WAVE_LENGTH,
    WAVE_NUMBER,
    check_and_adjust_instrument_name,
    convert2str,
    convert2wavenumber,
    download_rsr,
    get_bandname_from_wavelength,
    get_central_wave,
)

LOG = logging.getLogger(__name__)

OSCAR_PLATFORM_NAMES = {'eos-2': 'EOS-Aqua',
                        'meteosat-11': 'Meteosat-11',
                        'meteosat-10': 'Meteosat-10',
                        'meteosat-9': 'Meteosat-9',
                        'meteosat-8': 'Meteosat-8'}


class RSRDict(dict):
    """Helper dict-like class to handle multiple names for band keys."""

    def __init__(self, instrument=None):
        """Initialize dict and primary instrument name."""
        self.instrument = instrument
        dict.__init__(self)

    def __getitem__(self, key):
        """Get value either directly or fallback to pre-configured 'standard' names.."""
        try:
            val = dict.__getitem__(self, key)
        except KeyError:
            if self.instrument in BANDNAMES and key in BANDNAMES[self.instrument]:
                val = dict.__getitem__(self, BANDNAMES[self.instrument][key])
            elif key in BANDNAMES['generic']:
                val = dict.__getitem__(self, BANDNAMES['generic'][key])
            else:
                raise KeyError(f'Band not found in RSR for {self.instrument}: {key}')
        return val


class RSRDataBaseClass(object):
    """Data container for the Relative Spectral Responses for all (supported) satellite sensors."""

    def __init__(self):
        """Initialize the class instance.

        Create the instance either from platform name and instrument or from
        filename and load the data.

        """
        self.rsr = RSRDict()
        self.description = "Unknown"
        self.band_names = None
        self.unit = '1e-6 m'
        self.si_scale = 1e-6  # How to scale the wavelengths to become SI unit
        self._wavespace = WAVE_LENGTH

        options = get_config()
        self.rsr_dir = options['rsr_dir']
        self.do_download = False
        self._rsr_data_version_uptodate = False

        if 'download_from_internet' in options and options['download_from_internet']:
            self.do_download = True

        if self._get_rsr_data_version() == RSR_DATA_VERSION:
            self._rsr_data_version_uptodate = True

    def _get_rsr_data_version(self):
        """Check the version of the RSR data from the version file in the RSR directory."""
        rsr_data_version_path = os.path.join(self.rsr_dir, RSR_DATA_VERSION_FILENAME)
        if not os.path.exists(rsr_data_version_path):
            return "v0.0.0"

        with open(rsr_data_version_path, 'r') as fpt:
            # Get the version from the file
            return fpt.readline().strip()

    @property
    def rsr_data_version_uptodate(self):
        """Check whether RSR data are up to date."""
        return self._rsr_data_version_uptodate


class RelativeSpectralResponse(RSRDataBaseClass):
    """Container for the relative spectral response functions for various satellite imagers."""

    def __init__(self, platform_name=None, instrument=None, filename=None):
        """Initialize the class instance.

        Create the instance either from platform name and instrument or from
        filename, and then load the data from file.

        """
        super(RelativeSpectralResponse, self).__init__()

        self.platform_name = platform_name
        self.instrument = instrument
        self.filename = filename

        self._check_consistent_input()

        self.load()
        self.rsr.instrument = self.instrument

    def _check_consistent_input(self):
        """Check consistent input concerning platform name, instrument and RSR file name."""
        if not self.filename:
            if not self.instrument or not self.platform_name:
                raise AttributeError(
                    "Either platform name and sensor, or filename, must be specified")

            self._check_instrument()
            self._get_filename()

        self._check_filename_exist()

    def _check_instrument(self):
        """Check and try fix instrument name if needed."""
        self.instrument = check_and_adjust_instrument_name(self.platform_name, self.instrument)

    def _get_filename(self):
        """Get the rsr filname from platform and instrument names, and download if not available."""
        self.filename = expanduser(
            os.path.join(self.rsr_dir, 'rsr_{0}_{1}.h5'.format(self.instrument,
                                                               self.platform_name)))

        LOG.debug('Filename: %s', str(self.filename))
        if not os.path.exists(self.filename) or not os.path.isfile(self.filename):
            LOG.warning("No rsr file %s on disk", self.filename)

            if self._rsr_data_version_uptodate:
                LOG.info("RSR data up to date, so seems there is no support for this platform and sensor")
            else:
                # Try download from the internet!
                if self.do_download:
                    LOG.info("Will download from internet...")
                    download_rsr()
                    if self._get_rsr_data_version() == RSR_DATA_VERSION:
                        self._rsr_data_version_uptodate = True

        if not self._rsr_data_version_uptodate:
            LOG.warning("rsr data may not be up to date: %s", self.filename)
            if self.do_download:
                LOG.info("Will download from internet...")
                download_rsr()

    def _check_filename_exist(self):
        """Check that the file exist."""
        if not os.path.exists(self.filename) or not os.path.isfile(self.filename):
            errmsg = ('pyspectral RSR file does not exist! Filename = ' +
                      str(self.filename))
            if self.instrument and self.platform_name:
                fmatch = glob(
                    os.path.join(self.rsr_dir, '*{0}*{1}*.h5'.format(self.instrument,
                                                                     self.platform_name)))
                errmsg = (errmsg +
                          '\nFiles matching instrument and satellite platform' +
                          ': ' + str(fmatch))

            raise IOError(errmsg)

        LOG.debug('Filename: %s', str(self.filename))

    def load(self):
        """Read the internally formatet hdf5 relative spectral response data."""
        import h5py

        with h5py.File(self.filename, 'r') as h5f:
            self.set_band_names(h5f)
            self.set_description(h5f)
            self.set_platform_name(h5f)
            self.set_instrument(h5f)
            self.get_relative_spectral_responses(h5f)

    def integral(self, bandname):
        """Calculate the integral of the spectral response function for each detector."""
        intg = {}
        for det in self.rsr[bandname].keys():
            wvl = self.rsr[bandname][det]['wavelength']
            resp = self.rsr[bandname][det]['response']
            intg[det] = np.trapz(resp, wvl)
        return intg

    def convert(self):
        """Convert spectral response functions from wavelength to wavenumber."""
        if self._wavespace == WAVE_LENGTH:
            rsr, info = convert2wavenumber(self.rsr)
            for band in rsr.keys():
                for det in rsr[band].keys():
                    self.rsr[band][det][WAVE_NUMBER] = rsr[
                        band][det][WAVE_NUMBER]
                    self.rsr[band][det]['response'] = rsr[
                        band][det]['response']
                    self.unit = info['unit']
                    self.si_scale = info['si_scale']
            self._wavespace = WAVE_NUMBER
            for band in rsr.keys():
                for det in rsr[band].keys():
                    self.rsr[band][det]['central_wavenumber'] = \
                        get_central_wave(self.rsr[band][det][WAVE_NUMBER], self.rsr[band][det]['response'])
                    del self.rsr[band][det][WAVE_LENGTH]
        else:
            errmsg = "Conversion from {wn} to {wl} not supported yet".format(wn=WAVE_NUMBER, wl=WAVE_LENGTH)
            raise NotImplementedError(errmsg)

    def set_description(self, h5f):
        """Set the description."""
        self.description = h5f.attrs['description']
        self.description = convert2str(self.description)

    def set_band_names(self, h5f):
        """Set the band names."""
        self.band_names = h5f.attrs['band_names']
        self.band_names = [convert2str(x) for x in self.band_names]

    def set_instrument(self, h5f):
        """Set the instrument name."""
        if self.instrument:
            return

        try:
            self.instrument = h5f.attrs['sensor']
            self.instrument = convert2str(self.instrument)
        except KeyError:
            LOG.warning("No sensor name specified in HDF5 file")
            self.instrument = INSTRUMENTS.get(self.platform_name)

    def set_platform_name(self, h5f):
        """Set the platform name."""
        if self.platform_name:
            return

        try:
            self.platform_name = h5f.attrs['platform_name']
            self.platform_name = convert2str(self.platform_name)
        except KeyError:
            LOG.warning("No platform_name in HDF5 file")
            try:
                satname = h5f.attrs['platform']
                satname = convert2str(satname)
                sat_number = h5f.attrs['sat_number']
                self.platform_name = satname + '-' + str(sat_number)
            except KeyError:
                LOG.warning(
                    "Unable to determine platform name from HDF5 file content")
                self.platform_name = None

        self.platform_name = OSCAR_PLATFORM_NAMES.get(self.platform_name, self.platform_name)

    def get_number_of_detectors4bandname(self, h5f, bandname):
        """For a band name get the number of detectors, if any."""
        try:
            num_of_det = h5f[bandname].attrs['number_of_detectors']
        except KeyError:
            LOG.debug("No detectors found - assume only one...")
            num_of_det = 1

        return num_of_det

    def set_band_responses_per_detector(self, h5f, bandname, detector_name):
        """Set the RSR responses for the band and detector."""
        self.rsr[bandname][detector_name] = {}
        try:
            resp = h5f[bandname][detector_name]['response'][:]
        except KeyError:
            resp = h5f[bandname]['response'][:]

        self.rsr[bandname][detector_name]['response'] = resp

    def set_band_wavelengths_per_detector(self, h5f, bandname, detector_name):
        """Set the RSR wavelengths for the band and detector."""
        try:
            wvl = (h5f[bandname][detector_name]['wavelength'][:] *
                   h5f[bandname][detector_name]['wavelength'].attrs['scale'])
        except KeyError:
            wvl = (h5f[bandname]['wavelength'][:] *
                   h5f[bandname]['wavelength'].attrs['scale'])

        # The wavelength is given in micro meters!
        self.rsr[bandname][detector_name]['wavelength'] = wvl * 1e6

    def set_band_central_wavelength_per_detector(self, h5f, bandname, detector_name):
        """Set the central wavelength for the band and detector."""
        try:
            central_wvl = h5f[bandname][detector_name].attrs['central_wavelength']
        except KeyError:
            central_wvl = h5f[bandname].attrs['central_wavelength']

        self.rsr[bandname][detector_name]['central_wavelength'] = central_wvl

    def get_relative_spectral_responses(self, h5f):
        """Read the rsr data and add to the object."""
        for bandname in self.band_names:
            self.rsr[bandname] = {}

            num_of_det = self.get_number_of_detectors4bandname(h5f, bandname)
            for i in range(1, num_of_det + 1):
                dname = 'det-{0:d}'.format(i)
                self.set_band_responses_per_detector(h5f, bandname, dname)
                self.set_band_wavelengths_per_detector(h5f, bandname, dname)
                self.set_band_central_wavelength_per_detector(h5f, bandname, dname)

    def get_bandname_from_wavelength(self, wavelength, epsilon=0.1, multiple_bands=False):
        """Get the band name from the wavelength."""
        return get_bandname_from_wavelength(self.instrument, wavelength, self.rsr,
                                            epsilon=epsilon, multiple_bands=multiple_bands)


def check_and_download(dest_dir=None, dry_run=False):
    """Do a check for the version and attempt downloading only if needed."""
    rsr = RSRDataBaseClass()
    if rsr.rsr_data_version_uptodate:
        LOG.info("RSR data already the latest!")
    else:
        download_rsr(dest_dir=dest_dir, dry_run=dry_run)


if __name__ == "__main__":
    modis = RelativeSpectralResponse('EOS-Terra', 'modis')
    del modis



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Retrieve channel wavelength ranges for a given sensor."""

import argparse

from pyspectral.rsr_reader import RelativeSpectralResponse
from pyspectral.utils import get_wave_range


def _setup_argparse():
    arg_parser = argparse.ArgumentParser(
        description='Retrieve channel wavelength ranges for a given sensor.')
    arg_parser.add_argument("platform_name",
                            help="The Platform name",
                            type=str)
    arg_parser.add_argument("sensor",
                            help="The sensor/instrument name",
                            type=str)
    arg_parser.add_argument("-r", "--minimum_response",
                            help=("Minimum response threshold: Defines the value "
                                  "that the RSR must exceed in order to trigger"
                                  "the minimum or maximum wavelengths."),
                            default=0.15, type=float)
    arg_parser.add_argument("--bandname", '-b',
                            help="The sensor band name. Leave blank to return all bands.", type=str)
    arg_parser.add_argument("--detector", '-d',
                            help="The sensor detector, if not passed will default to det-1",
                            default='det-1', type=str)
    return arg_parser


def main(arguments):
    """Retrieve wavelength range based on user-supplied arguments."""
    platform = arguments.platform_name
    sensor = arguments.sensor
    threshold = arguments.minimum_response

    rsr = RelativeSpectralResponse(platform, sensor)
    if arguments.bandname:
        bands = [arguments.bandname]
    else:
        bands = rsr.band_names
    det = arguments.detector

    for bname in bands:
        wvls = get_wave_range(rsr.rsr[bname][det], threshold)
        print(f'name:  {bname}')
        print(f'  wavelength: [{wvls[0]:5.3f}, {wvls[1]:5.3f}, {wvls[2]:5.3f}]')


if __name__ == "__main__":

    parser = _setup_argparse()
    args = parser.parse_args()
    main(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the solar_flux calculations."""

import os
import unittest

import numpy as np

from pyspectral.solar import SolarIrradianceSpectrum

TEST_RSR = {}
TEST_RSR['det-1'] = {}
TEST_RSR['det-1']['wavelength'] = np.array([
    3.6123999, 3.6163599, 3.6264927, 3.6363862, 3.646468,
    3.6564937, 3.6664478, 3.6765388, 3.6865413, 3.6964585,
    3.7065142, 3.716509, 3.7264658, 3.7364102, 3.7463682,
    3.7563652, 3.7664226, 3.7763396, 3.7863384, 3.7964207,
    3.8063589, 3.8163606, 3.8264089, 3.8364836, 3.8463381,
    3.8563975, 3.8664163, 3.8763755, 3.8864797, 3.8964978,
    3.9064275, 3.9164873, 3.9264729, 3.9364026, 3.9465107,
    3.9535347], dtype='double')

TEST_RSR['det-1']['response'] = np.array([
    0.01, 0.0118, 0.01987, 0.03226, 0.05028, 0.0849,
    0.16645, 0.33792, 0.59106, 0.81815, 0.96077, 0.92855,
    0.86008, 0.8661, 0.87697, 0.85412, 0.88922, 0.9541,
    0.95687, 0.91037, 0.91058, 0.94256, 0.94719, 0.94808,
    1., 0.92676, 0.67429, 0.44715, 0.27762, 0.14852,
    0.07141, 0.04151, 0.02925, 0.02085, 0.01414, 0.01], dtype='double')


RESULT_IPOL_WVLS = np.array([0.2, 0.201, 0.202, 0.203, 0.204, 0.205, 0.206, 0.207, 0.208,
                             0.209, 0.21, 0.211, 0.212, 0.213, 0.214, 0.215, 0.216, 0.217,
                             0.218, 0.219, 0.22, 0.221, 0.222, 0.223, 0.224, 0.225, 0.226,
                             0.227, 0.228, 0.229, 0.23, 0.231, 0.232, 0.233, 0.234, 0.235,
                             0.236, 0.237, 0.238, 0.239, 0.24], dtype='double')


class TestSolarflux(unittest.TestCase):
    """Unit testing the solar flux calculations."""

    def setUp(self):
        """Set up."""
        self.solar_irr = SolarIrradianceSpectrum(dlambda=0.005)
        self.rsr = TEST_RSR

    def test_read(self):
        """Test that solar irradiance spctrum."""
        self.assertTrue(os.path.exists(self.solar_irr.filename))
        self.assertEqual(self.solar_irr.wavelength.shape[0], 1697)
        self.assertEqual(self.solar_irr.irradiance.shape[0], 1697)

    def test_solar_flux(self):
        """Calculate the solar-flux."""
        # rsr function (se above) is given in micronsm therefore the scale
        # factor is 1.0 and not 1e+6 (default)!
        sflux = self.solar_irr.inband_solarflux(self.rsr, scale=1.0)
        self.assertAlmostEqual(sflux, 2.002927627)
        # self.assertAlmostEqual(sflux, 2.5)

    def test_interpolate(self):
        """Test the interpolate method."""
        self.solar_irr.interpolate(dlambda=0.001, ival_wavelength=(0.200, 0.240))
        self.assertTrue(np.allclose(RESULT_IPOL_WVLS, self.solar_irr.ipol_wavelength))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Base class for reading raw instrument spectral responses."""

import logging
import os
from pathlib import Path

from pyspectral.config import get_config

LOG = logging.getLogger(__name__)


class InstrumentRSR(object):
    """Base class for the raw (agency dependent) instrument response functions."""

    def __init__(self, bandname, platform_name, bandnames=None):
        """Initialize the class instance."""
        self.platform_name = platform_name
        self.bandname = bandname
        self.instrument = None
        self.filenames = {}
        self.requested_band_filename = None
        self.bandnames = bandnames
        if not self.bandnames:
            self.bandnames = []
        for band in self.bandnames:
            self.filenames[band] = None
        self.rsr = None

        self.output_dir = None
        self.path = None
        self.filename = None
        self.options = {}

    def _get_options_from_config(self):
        """Get configuration settings from configuration file."""
        options = get_config()
        self.output_dir = options.get('rsr_dir', './')
        self.path = Path(options[self.platform_name + '-' + self.instrument]['path'])
        self.filename = options[self.platform_name + '-' + self.instrument].get('filename')
        self.options = options

    def _get_bandfilenames(self):
        """Get the instrument rsr filenames."""
        for band in self.bandnames:
            LOG.debug("Band = %s", str(band))
            self.filenames[band] = os.path.join(self.path,
                                                self.options[self.platform_name + '-' +
                                                             self.instrument][band])
            LOG.debug(self.filenames[band])
            if not os.path.exists(self.filenames[band]):
                LOG.warning("Couldn't find an existing file for this band: %s",
                            str(self.filenames[band]))

    def _load(self, scale=1.0):
        """Load the instrument RSR from file(s)."""
        raise NotImplementedError(
            "Instrument rsr loader needs to be defined in the subclass")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the generic rsr hdf5 reader."""
import contextlib
import os.path
import unittest
from unittest.mock import patch

import numpy as np
import pytest
import xarray as xr

from pyspectral.rsr_reader import RelativeSpectralResponse, RSRDict
from pyspectral.tests.unittest_helpers import assertNumpyArraysEqual
from pyspectral.utils import RSR_DATA_VERSION, RSR_DATA_VERSION_FILENAME, WAVE_NUMBER

TEST_RSR = {'20': {}, }
TEST_RSR['20']['det-1'] = {}
TEST_RSR['20']['det-1']['central_wavelength'] = 3.75
TEST_RSR['20']['det-1']['wavelength'] = np.array([
    3.6123999, 3.6163599, 3.6264927, 3.6363862, 3.646468,
    3.6564937, 3.6664478, 3.6765388, 3.6865413, 3.6964585,
    3.7065142, 3.716509, 3.7264658, 3.7364102, 3.7463682,
    3.7563652, 3.7664226, 3.7763396, 3.7863384, 3.7964207,
    3.8063589, 3.8163606, 3.8264089, 3.8364836, 3.8463381,
    3.8563975, 3.8664163, 3.8763755, 3.8864797, 3.8964978,
    3.9064275, 3.9164873, 3.9264729, 3.9364026, 3.9465107,
    3.9535347], dtype='float32')

TEST_RSR['20']['det-1']['response'] = np.array([
    0.01, 0.0118, 0.01987, 0.03226, 0.05028, 0.0849,
    0.16645, 0.33792, 0.59106, 0.81815, 0.96077, 0.92855,
    0.86008, 0.8661, 0.87697, 0.85412, 0.88922, 0.9541,
    0.95687, 0.91037, 0.91058, 0.94256, 0.94719, 0.94808,
    1., 0.92676, 0.67429, 0.44715, 0.27762, 0.14852,
    0.07141, 0.04151, 0.02925, 0.02085, 0.01414, 0.01], dtype='float32')

TEST_RSR2 = {'20': {}, }
TEST_RSR2['20']['det-1'] = {}
TEST_RSR2['20']['det-1']['central_wavelength'] = 3.75
TEST_RSR2['20']['det-1']['wavelength'] = TEST_RSR['20']['det-1']['wavelength'].copy()
TEST_RSR2['20']['det-1']['response'] = TEST_RSR['20']['det-1']['response'].copy()

RESULT_WVN_RSR = np.array([2529.38232422,  2533.8840332,  2540.390625,  2546.81494141,
                           2553.30859375,  2559.88378906,  2566.40722656,  2573.02270508,
                           2579.72949219,  2586.37451172,  2593.09375,  2599.87548828,
                           2606.55371094,  2613.41674805,  2620.29760742,  2627.18286133,
                           2634.06005859,  2641.07421875,  2648.06713867,  2655.03930664,
                           2662.14794922,  2669.25170898,  2676.36572266,  2683.50805664,
                           2690.69702148,  2697.95288086,  2705.29199219,  2712.56982422,
                           2719.94970703,  2727.43554688,  2734.8605957,  2742.37988281,
                           2749.9831543,  2757.48510742,  2765.21142578,  2768.24291992], dtype=np.float32)

DIR_PATH_ITEMS = ['test', 'path', 'to', 'rsr', 'data']
TEST_CONFIG = {}

TEST_RSR_DIR = os.path.join(*DIR_PATH_ITEMS)
TEST_CONFIG['rsr_dir'] = TEST_RSR_DIR


class TestRsrReader(unittest.TestCase):
    """Class for testing pyspectral.rsr_reader."""

    @patch('os.path.exists')
    @patch('os.path.isfile')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.download_rsr')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._get_rsr_data_version')
    def test_get_rsr_from_platform_name_and_instrument(self, get_rsr_version,
                                                       download_rsr, load, isfile, exists):
        """Test the RelativeSpectralResponse class initialisation.

        Test getting the rsr from platform_name and instrument.
        """
        load.return_code = None
        download_rsr.return_code = None
        isfile.return_code = True
        exists.return_code = True
        get_rsr_version.return_code = RSR_DATA_VERSION

        with patch('pyspectral.rsr_reader.get_config', return_value=TEST_CONFIG):
            with self.assertRaises(AttributeError):
                _ = RelativeSpectralResponse('GOES-16')

            with self.assertRaises(AttributeError):
                _ = RelativeSpectralResponse(instrument='ABI')

    @patch('os.path.exists')
    @patch('os.path.isfile')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.download_rsr')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._get_rsr_data_version')
    def test_convert(self, get_rsr_version, download_rsr, load, isfile, exists):
        """Test the conversion method."""
        load.return_code = None
        download_rsr.return_code = None
        isfile.return_code = True
        exists.return_code = True
        get_rsr_version.return_code = RSR_DATA_VERSION

        with patch('pyspectral.rsr_reader.get_config', return_value=TEST_CONFIG):
            test_rsr = RelativeSpectralResponse('EOS-Aqua', 'modis')
            test_rsr.rsr = TEST_RSR
            test_rsr.convert()
            self.assertAlmostEqual(test_rsr.rsr['20']['det-1']['central_wavenumber'], 2647.397, 3)
            self.assertTrue(np.allclose(test_rsr.rsr['20']['det-1'][WAVE_NUMBER], RESULT_WVN_RSR, 5))
            self.assertEqual(test_rsr._wavespace, WAVE_NUMBER)

            with self.assertRaises(NotImplementedError):
                test_rsr.convert()

    @patch('os.path.exists')
    @patch('os.path.isfile')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.download_rsr')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._get_rsr_data_version')
    def test_integral(self, get_rsr_version, download_rsr, load, isfile, exists):
        """Test the calculation of the integral of the spectral responses."""
        load.return_code = None
        download_rsr.return_code = None
        isfile.return_code = True
        exists.return_code = True
        get_rsr_version.return_code = RSR_DATA_VERSION

        with patch('pyspectral.rsr_reader.get_config', return_value=TEST_CONFIG):
            test_rsr = RelativeSpectralResponse('EOS-Aqua', 'modis')
            test_rsr.rsr = TEST_RSR2
            res = test_rsr.integral('20')
            self.assertAlmostEqual(res['det-1'], 0.185634, 6)


class MyHdf5Mock(object):
    """A Mock for the RSR data normally stored in a HDF5 file."""

    def __init__(self, attrs):
        """Initialize the mock class."""
        self.attrs = attrs


class TestPopulateRSRObject(unittest.TestCase):
    """Testing populate the RelativeSpectralResponse instance from the hdf5 file."""

    def setUp(self):
        """Initialise the tests."""
        bandnames = np.array([b'1', b'2', b'3', b'4', b'5', b'6', b'7', b'8', b'9', b'10', b'11',
                              b'12', b'13', b'14', b'15', b'16', b'17', b'18', b'19', b'20',
                              b'21', b'22', b'23', b'24', b'25', b'26', b'27', b'28', b'29',
                              b'30', b'31', b'32', b'33', b'34', b'35', b'36'], dtype='|S2')
        hdf5_attrs_aqua_modis = {'band_names': bandnames,
                                 'description': 'Relative Spectral Responses for MODIS',
                                 'platform': 'eos',
                                 'sat_number': 2}
        hdf5_attrs_aqua_modis_b = {'band_names': bandnames,
                                   'description': b'Relative Spectral Responses for MODIS',
                                   'platform': b'eos',
                                   'sat_number': 2}
        self.modis_bandnames = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10',
                                '11', '12', '13', '14', '15', '16', '17', '18', '19', '20',
                                '21', '22', '23', '24', '25', '26', '27', '28', '29', '30',
                                '31', '32', '33', '34', '35', '36']

        self.h5f_aqua_modis = MyHdf5Mock(hdf5_attrs_aqua_modis)
        self.h5f_aqua_modis_b = MyHdf5Mock(hdf5_attrs_aqua_modis_b)

        hdf5_attrs_aqua_modis_err = {'band_names': bandnames,
                                     'description': 'Relative Spectral Responses for MODIS',
                                     'platform': 'eos'}

        self.h5f_aqua_modis_err = MyHdf5Mock(hdf5_attrs_aqua_modis_err)

        bandnames = np.array([b'M1', b'M2', b'M3', b'M4', b'M5', b'M6', b'M7', b'M8', b'M9',
                              b'M10', b'M11', b'M12', b'M13', b'M14', b'M15', b'M16', b'I1',
                              b'I2', b'I3', b'I4', b'I5', b'DNB'], dtype='|S3')

        hdf5_attrs_noaa20_viirs = {'band_names': bandnames,
                                   'description': b'Relative Spectral Responses for VIIRS',
                                   'platform_name': b'NOAA-20',
                                   'sensor': b'viirs'}
        self.viirs_bandnames = ['M1', 'M2', 'M3', 'M4', 'M5', 'M6', 'M7', 'M8', 'M9', 'M10',
                                'M11', 'M12', 'M13', 'M14', 'M15', 'M16',
                                'I1', 'I2', 'I3', 'I4', 'I5', 'DNB']
        self.viirs_detector_names = ['det-1', 'det-2', 'det-3', 'det-4', 'det-5',
                                     'det-6', 'det-7', 'det-8', 'det-9', 'det-10',
                                     'det-11', 'det-12', 'det-13', 'det-14', 'det-15',
                                     'det-16']
        self.viirs_rsr_data = xr.Dataset({'wavelength': xr.DataArray(np.arange(10)),
                                          'response': xr.DataArray(np.arange(10))})
        self.viirs_rsr_data.attrs['central_wavelength'] = 100.0
        self.viirs_rsr_data['wavelength'].attrs['scale'] = 0.01

        self.h5f_noaa20_viirs = MyHdf5Mock(hdf5_attrs_noaa20_viirs)

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_filename_exist')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._get_filename')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_instrument')
    def test_create_rsr_instance(self, check_instrument, get_filename, check_filename_exist, load):
        """Test creating the instance."""
        load.return_value = None
        check_filename_exist.return_value = None
        get_filename.return_value = None
        check_instrument.return_value = None

        with pytest.raises(AttributeError) as exec_info:
            _ = RelativeSpectralResponse('MyPlatform')

        exception_raised = exec_info.value
        expected_value = 'Either platform name and sensor, or filename, must be specified'
        self.assertEqual(str(exception_raised), expected_value)

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_consistent_input')
    def test_set_description(self, check_consistent_input, load):
        """Test setting the description."""
        load.return_value = None
        check_consistent_input.return_value = None

        test_rsr = RelativeSpectralResponse('NOAA-20', 'viirs')
        test_rsr.set_description(self.h5f_noaa20_viirs)

        self.assertEqual(test_rsr.description, 'Relative Spectral Responses for VIIRS')

        test_rsr = RelativeSpectralResponse('EOS-Aqua', 'modis')
        test_rsr.set_description(self.h5f_aqua_modis)

        self.assertEqual(test_rsr.description, 'Relative Spectral Responses for MODIS')

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_consistent_input')
    def test_set_platform_name(self, check_consistent_input, load):
        """Test setting the platform name."""
        load.return_value = None
        check_consistent_input.return_value = None

        test_rsr = RelativeSpectralResponse('EOS-Aqua', 'modis')
        test_rsr.set_platform_name(self.h5f_aqua_modis)
        self.assertEqual(test_rsr.platform_name, 'EOS-Aqua')

        test_rsr = RelativeSpectralResponse('MyPlatform', 'modis')
        self.assertEqual(test_rsr.platform_name, 'MyPlatform')

        modis_aqua_filepath = "/mypath/myfilename"
        test_rsr = RelativeSpectralResponse(filename=modis_aqua_filepath)
        test_rsr.filename = modis_aqua_filepath

        test_rsr.set_platform_name(self.h5f_aqua_modis)
        self.assertEqual(test_rsr.platform_name, 'EOS-Aqua')

        modis_aqua_filepath = "/mypath/myfilename"
        test_rsr = RelativeSpectralResponse(filename=modis_aqua_filepath)
        test_rsr.filename = modis_aqua_filepath

        test_rsr.set_platform_name(self.h5f_aqua_modis_b)
        self.assertEqual(test_rsr.platform_name, 'EOS-Aqua')

        viirs_noaa20_filepath = "/mypath/myfilename"
        test_rsr = RelativeSpectralResponse(filename=viirs_noaa20_filepath)
        test_rsr.filename = viirs_noaa20_filepath

        test_rsr.set_platform_name(self.h5f_noaa20_viirs)
        self.assertEqual(test_rsr.platform_name, 'NOAA-20')

        test_rsr = RelativeSpectralResponse(filename=modis_aqua_filepath)
        test_rsr.filename = modis_aqua_filepath
        test_rsr.set_platform_name(self.h5f_aqua_modis_err)
        self.assertEqual(test_rsr.platform_name, None)

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_consistent_input')
    def test_set_instrument(self, check_consistent_input, load):
        """Test setting the instrument name."""
        load.return_value = None
        check_consistent_input.return_value = None

        test_rsr = RelativeSpectralResponse('EOS-Aqua', 'modis')
        test_rsr.set_instrument(self.h5f_aqua_modis)
        self.assertEqual(test_rsr.instrument, 'modis')

        modis_aqua_filepath = "/mypath/myfilename"
        test_rsr = RelativeSpectralResponse(filename=modis_aqua_filepath)
        test_rsr.filename = modis_aqua_filepath
        test_rsr.platform_name = 'EOS-Aqua'
        test_rsr.set_instrument(self.h5f_aqua_modis)
        self.assertEqual(test_rsr.instrument, 'modis')

        viirs_noaa20_filepath = "/mypath/myfilename"
        test_rsr = RelativeSpectralResponse(filename=viirs_noaa20_filepath)
        test_rsr.filename = viirs_noaa20_filepath

        test_rsr.set_instrument(self.h5f_noaa20_viirs)
        self.assertEqual(test_rsr.instrument, 'viirs')

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_consistent_input')
    def test_set_band_names(self, check_consistent_input, load):
        """Test setting the band names."""
        load.return_value = None
        check_consistent_input.return_value = None

        test_rsr = RelativeSpectralResponse('EOS-Aqua', 'modis')
        test_rsr.set_band_names(self.h5f_aqua_modis)
        expected = self.modis_bandnames
        self.assertCountEqual(test_rsr.band_names, expected)

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_consistent_input')
    def test_set_band_central_wavelength_per_detector(self,
                                                      check_consistent_input, load):
        """Test setting the band specific central wavelength for a detector."""
        load.return_value = None
        check_consistent_input.return_value = None

        test_rsr = RelativeSpectralResponse('NOAA-20', 'viirs')

        h5f = {}
        for name in self.viirs_bandnames:
            h5f[name] = {}
            for det_name in self.viirs_detector_names:
                h5f[name][det_name] = self.viirs_rsr_data

        test_rsr.rsr['M1'] = {'det-1': {}}
        test_rsr.set_band_central_wavelength_per_detector(h5f, 'M1', 'det-1')
        self.assertIn('central_wavelength', test_rsr.rsr['M1']['det-1'])
        self.assertEqual(test_rsr.rsr['M1']['det-1']['central_wavelength'], 100.)

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_consistent_input')
    def test_set_band_wavelengths_per_detector(self,
                                               check_consistent_input, load):
        """Test setting the band wavelengths for a detector."""
        load.return_value = None
        check_consistent_input.return_value = None

        test_rsr = RelativeSpectralResponse('NOAA-20', 'viirs')

        h5f = {}
        for name in self.viirs_bandnames:
            h5f[name] = {}
            for det_name in self.viirs_detector_names:
                h5f[name][det_name] = self.viirs_rsr_data

        test_rsr.rsr['M1'] = {'det-1': {}}
        test_rsr.set_band_wavelengths_per_detector(h5f, 'M1', 'det-1')

        expected = np.array([0., 10000., 20000., 30000., 40000., 50000.,
                             60000., 70000., 80000., 90000.])
        assertNumpyArraysEqual(test_rsr.rsr['M1']['det-1']['wavelength'].data, expected)

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_consistent_input')
    def test_set_band_responses_per_detector(self,
                                             check_consistent_input, load):
        """Test setting the band responses for a detector."""
        load.return_value = None
        check_consistent_input.return_value = None

        test_rsr = RelativeSpectralResponse('NOAA-20', 'viirs')

        h5f = {}
        for name in self.viirs_bandnames:
            h5f[name] = {}
            for det_name in self.viirs_detector_names:
                h5f[name][det_name] = self.viirs_rsr_data

        test_rsr.rsr['M1'] = {'det-1': {}}
        test_rsr.set_band_responses_per_detector(h5f, 'M1', 'det-1')

        expected = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        assertNumpyArraysEqual(test_rsr.rsr['M1']['det-1']['response'].data, expected)

    @patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
    @patch('pyspectral.rsr_reader.RelativeSpectralResponse._check_consistent_input')
    def test_get_bandname_from_rsr(self,
                                   check_consistent_input, load):
        """Test get the matching band name from the rsr object."""
        load.return_value = None
        check_consistent_input.return_value = None

        test_rsr = RelativeSpectralResponse('EOS-Aqua', 'modis')
        test_rsr.rsr = TEST_RSR

        myband = test_rsr.get_bandname_from_wavelength(3.75)
        assert myband == '20'

    def test_rsr_dict(self):
        """Test finding correct band names from utils dicts."""
        test_rsr = RSRDict(instrument='viirs')
        test_rsr['M1'] = 0
        test_rsr['VIS0.6'] = 1
        # Check for correct band name
        self.assertEqual(test_rsr['M1'], 0)
        # Check for alternative band name on same instrument
        self.assertEqual(test_rsr['M01'], 0)
        # Check for generic band name
        self.assertEqual(test_rsr['VIS006'], 1)
        # Check exception raised if incorrect band name given
        with self.assertRaises(KeyError):
            print('d', test_rsr['VIS030'])

    def test_rsr_unconfigured_sensor(self):
        """Test RSRDict finds generic band conversions when specific sensor is not configured."""
        test_rsr = RSRDict(instrument="i dont exist")
        test_rsr["ch1"] = 2
        assert test_rsr['1'] == 2


@patch('os.path.exists')
@patch('os.path.isfile')
@patch('pyspectral.rsr_reader.RelativeSpectralResponse.load')
@patch('pyspectral.rsr_reader.download_rsr')
@patch('pyspectral.rsr_reader.RelativeSpectralResponse._get_rsr_data_version')
@pytest.mark.parametrize(
    ("platform_name", "instrument", "exp_filename", "exp_instrument"),
    [
        ('FY-3D', 'mersi-2', 'rsr_mersi2_FY-3D.h5', 'mersi2'),
        ('TIROS-N', 'avhrr-1', 'rsr_avhrr1_TIROS-N.h5', 'avhrr1'),
        ('NOAA-12', 'avhrr-2', 'rsr_avhrr2_NOAA-12.h5', 'avhrr2'),
        ('NOAA-12', 'avhrr/2', 'rsr_avhrr2_NOAA-12.h5', 'avhrr2'),
        ('NOAA-19', 'avhrr/3', 'rsr_avhrr3_NOAA-19.h5', 'avhrr3'),
        ('GOES-16', 'abi', 'rsr_abi_GOES-16.h5', 'abi'),
        ('GOES-16', 'ABI', 'rsr_abi_GOES-16.h5', 'abi'),
        ('GOES-18', 'AbI', 'rsr_abi_GOES-18.h5', 'abi'),
    ]
)
def test_get_rsr_from_platform_and_instrument(get_rsr_version,
                                              download_rsr, load, isfile, exists,
                                              platform_name, instrument,
                                              exp_filename, exp_instrument):
    """Test getting the rsr filename correct when specifying the platform and instrument names."""
    load.return_code = None
    download_rsr.return_code = None
    isfile.return_code = True
    exists.return_code = True
    get_rsr_version.return_code = RSR_DATA_VERSION

    with patch('pyspectral.rsr_reader.get_config', return_value=TEST_CONFIG):
        test_rsr = RelativeSpectralResponse(platform_name, instrument)
        assert test_rsr.platform_name == platform_name
        assert test_rsr.instrument == exp_instrument
        assert os.path.basename(test_rsr.filename) == exp_filename

    # def test_get_bandname_from_wavelength_inconsistent_instrument_name(self):
    #     """Test the right bandname is found provided the wavelength in micro meters.

    #     Here we test that if you enter with a sensor name different from what
    #     is set in the RSR object, an error is raised.

    #     """
    #     with pytest.raises(AttributeError):
    #         bname = utils.get_bandname_from_wavelength('ahi', 0.4, self.rsr)


@pytest.mark.parametrize(
    ("version", "exp_download"),
    [
        (RSR_DATA_VERSION, False),
        ("v1.0.0", True),
    ],
)
def test_check_and_download(tmp_path, version, exp_download):
    """Test that check_and_download only downloads when necessary."""
    from pyspectral.rsr_reader import check_and_download
    with _fake_rsr_dir(tmp_path, version), unittest.mock.patch("pyspectral.rsr_reader.download_rsr") as download:
        check_and_download()
        if exp_download:
            download.assert_called()
        else:
            download.assert_not_called()


@contextlib.contextmanager
def _fake_rsr_dir(tmp_path, rsr_version):
    with _fake_get_config(tmp_path):
        version_filename = str(tmp_path / RSR_DATA_VERSION_FILENAME)
        with open(version_filename, "w") as version_file:
            version_file.write(rsr_version)
        yield


@contextlib.contextmanager
def _fake_get_config(tmp_path):
    def _get_config():
        return {
            "rayleigh_dir": str(tmp_path),
            "rsr_dir": str(tmp_path),
            "download_from_internet": True,
        }
    with unittest.mock.patch("pyspectral.rsr_reader.get_config") as get_config:
        get_config.side_effect = _get_config
        yield
